/*	File			DB_Arch_DBServer_DB1_Direct.java
	Purpose			DB SERVER component for illustration of aspects of connectivity, coupling, architecture and transparency
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			March 2015
*/
package DB_Arch_DBServer_DB1_Direct;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.*;
    
public class DB_Arch_DBServer_DB1_Direct extends javax.swing.JFrame implements ActionListener
{
    private static int DB_TYPE1_PORT = 8002;
    private static int DB_TYPE2_PORT = 8003;
    private static int DATABASE_SIZE = 3;
    private static int MAX_CUST_NAME_LEN = 20;
    
    private enum MessageType {REQUEST_RECORD, SEND_RECORD, CLIENT_DISCONNECT};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Application_Message_PDU
    {
        MessageType eMessageType;   // Size 4   Offset 0
        int iCust_ID;               // Size 4   Offset 4
        Boolean bCustNameValid;     // Size 1   Offset 8
        byte[] cCustName;           // Size 21  Offset 9 - 29
        Boolean bCustBalanceValid;  // Size 1   Offset 30
        // Single byte of padding exists in the byte arraay at this point, to align the following int on a 4-byte boundary
        int iCustBalance;           // Size 32  Offset 32
    };

    private static int Application_Message_Size = 36;
    
    ServerSocket m_ListenSock;  // Listen for client connections;
    InetAddress m_Local_Address;
    Socket m_ConnectSock;       // Connection to Client
    InputStream m_InStream;
    OutputStream m_OutStream;
    InetAddress m_Connect_Address;
    Boolean m_bConnected;
    Timer m_Timer;

    // Data Row and Table structures (hard coded data serves as mock of real database)
    final class DataRow_struct 
    {
        int iCust_ID;
        byte[] cCustName;	
    };

    // DataTable_struct
    DataRow_struct[] m_DataRowArray; // Array of data row structures
                 
    public DB_Arch_DBServer_DB1_Direct()
    {
        initComponents();
        GetLocalIPAddress();
        m_bConnected = false;
        InitialiseDatabase();
        DisplayDataTable();
        EnableServer();
        InitialiseControls();
    }

    private void InitialiseControls()    
    {    
        jTextArea_CustomerID.setEnabled(false);;
        jTextArea_CustomerName.setEnabled(false);
        jButton_ClearLog.setEnabled(true); 
        jButton_Done.setEnabled(true); 
        jTextArea_EventLog.setEnabled(false);
    }

    void GetLocalIPAddress()
    {   
        String sHostAddress = "";
        try
        {
            m_Local_Address = InetAddress.getLocalHost();
            sHostAddress = m_Local_Address.toString(); // Expected format: "Hostname/192.168.100.5"
            String sAddressDisplay = "Server address " + sHostAddress;
            WriteEventLogEntry(sAddressDisplay);            
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"DB Arch DBServer1",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
    }
    
    void InitialiseDatabase()	// Populate DataTable (hardcoded data)
    {
        m_DataRowArray = new DataRow_struct[DATABASE_SIZE];
        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {
            m_DataRowArray[iIndex] = new DataRow_struct();
            m_DataRowArray[iIndex].cCustName = new byte[MAX_CUST_NAME_LEN+1];
        }
        m_DataRowArray[0].iCust_ID = 101;
        String sCustName = "Fred";
        m_DataRowArray[0].cCustName = sCustName.getBytes();
        m_DataRowArray[1].iCust_ID = 102;
        sCustName = "Jane";
        m_DataRowArray[1].cCustName = sCustName.getBytes();
        m_DataRowArray[2].iCust_ID = 103;
        sCustName = "Rita";
        m_DataRowArray[2].cCustName = sCustName.getBytes();
    }

    void DisplayDataTable()
    {	
        jTextArea_CustomerID.setText("");
        jTextArea_CustomerName.setText("");
        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex ++)
        {
            jTextArea_CustomerID.append(Integer.toString(m_DataRowArray[iIndex].iCust_ID) + "\n");
            jTextArea_CustomerName.append(new String(m_DataRowArray[iIndex].cCustName) + "\n");
        }
    }

    void Accept()
    {
        try
        {
            m_ConnectSock = m_ListenSock.accept();
            m_ConnectSock.setSoTimeout(100); // 100ms timeout (non-blocking)            }
            m_InStream = m_ConnectSock.getInputStream();
            m_OutStream = m_ConnectSock.getOutputStream();
        }
        catch (SocketTimeoutException Ex)
        {   
            return;
        }
        catch (Exception Ex)
        {
            WriteEventLogEntry(Ex.toString());
            return;
        }

        m_bConnected = true;
        String sStr = String.format("Client %s Connected", m_ConnectSock.getRemoteSocketAddress().toString());
        WriteEventLogEntry(sStr);
    }

    void EnableServer()  // Invoked on startup
    {
        try
        {   // Create, Bind and Listen on Listen Socket and set it NonBlocking
            m_ListenSock = new ServerSocket(DB_TYPE1_PORT, 5/*backlog*/, m_Local_Address/*bindAddr*/);
            m_ListenSock.setSoTimeout(100); // Timeout = 100mS i.e. non-blocking
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"DB Arch DBServer1",0);
           CloseAndExit();
        }

        Start_Timer();   // As all activity is non-blocking
    }
    
    private void Start_Timer()
    {
        m_Timer = new Timer();
        m_Timer.scheduleAtFixedRate(
                new OnTimedEvent_Timer(), 100, 100); // Short interval to allow UI responsiveness)
    }

    private class OnTimedEvent_Timer extends TimerTask
    {
        public void run()
        {
            if(true == m_bConnected)
            {
                    DoReceive();
            }
            else
            {
                    Accept();
            }
        }
    }

    void SendApplicationMessage(Application_Message_PDU Message)
    {
        byte[] bBuffer = Serialize_Application_Message_PDU(Message);
        try
        {
            m_OutStream.write(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
        }
        catch (Exception ex)
        {
        }
    }

    void Send_DataRow_Message(int iCustID)
    {
        Application_Message_PDU Message = new Application_Message_PDU();
        Message.eMessageType = MessageType.SEND_RECORD;
        Message.iCust_ID = iCustID;
        String sCustName = "------No Match------";
        Message.cCustName = sCustName.getBytes();

        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {
            if(iCustID == m_DataRowArray[iIndex].iCust_ID)
            {
                Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
                System.arraycopy(m_DataRowArray[iIndex].cCustName/*src*/, 0, Message.cCustName/*dest*/, 0, m_DataRowArray[iIndex].cCustName.length/*count*/); 
            }
        }

        Message.bCustNameValid = true;
        Message.bCustBalanceValid = false;
        SendApplicationMessage(Message);
        String sCust = new String(Message.cCustName);
        sCust = TruncateString(sCust, sCust.length());        
        String sStr = String.format("Data Record %d %s sent", Message.iCust_ID, sCust);
        WriteEventLogEntry(sStr);
    }

    void DoReceive()
    {		
        byte[] bArray = new byte[Application_Message_Size];
        int iReceiveByteCount = 0;
        if (true == m_bConnected)
        {
            iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_InStream.read(bArray);
            }
            catch (SocketTimeoutException Ex)
            {
            }
            catch (Exception Ex)
            {
            }
            if(Application_Message_Size <= iReceiveByteCount)
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
                String sStr;
                switch(Message.eMessageType)
                {
                   case REQUEST_RECORD:
                        sStr = String.format("Data record %d requested", Message.iCust_ID);
                        WriteEventLogEntry(sStr);
                        Send_DataRow_Message(Message.iCust_ID);
                        break;		
                    case CLIENT_DISCONNECT:
                        m_bConnected = false;
                        CloseConnectionSocket();
                        WriteEventLogEntry("Client disconnected");
                        break;
                }
            }
        }
    }

    Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
        Application_Message_PDU result_Message = new Application_Message_PDU();
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iCust_ID = ByteArraySize4_ToInt(bInt);
        result_Message.bCustNameValid = true;
        if (0 == bRecdByteArray[8])
        {
            result_Message.bCustNameValid = false;
        }
        result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
        System.arraycopy(bRecdByteArray/*src*/, 9, result_Message.cCustName/*dest*/, 0, MAX_CUST_NAME_LEN/*count*/); 
        result_Message.bCustBalanceValid = true;
        if (0 == bRecdByteArray[30])
        {
            result_Message.bCustBalanceValid = false;
        }
        System.arraycopy(bRecdByteArray/*src*/, 32, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iCustBalance = ByteArraySize4_ToInt(bInt);
        return result_Message;
    }

    private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
    {   // Convert an Application_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Application_Message_Size]; // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iCust_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bByteArray[8] = 0; // false (bCustNameValid)
        if(true == Message.bCustNameValid)
        {
            bByteArray[8] = 1; // true (bCustNameValid)
        }
        if(null != Message.cCustName)
        {
            System.arraycopy(Message.cCustName/*src*/, 0, bByteArray/*dest*/, 9, Message.cCustName.length/*count*/);
        }
        bByteArray[30] = 0; // false (bCustBalanceValid)
        if (true == Message.bCustBalanceValid)
        {
            bByteArray[30] = 1; // true (bCustBalanceValid)
        }
        bInt = IntToByteArray(Message.iCustBalance);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 32, 4/*count*/);
        return bByteArray;
    }

    void CloseConnectionSocket()
    {
        try
        {
            m_ConnectSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }
    
    void CloseSockets()
    {
        CloseConnectionSocket();
        try
        {
            m_ListenSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
           if(null != m_Timer)
            {
                m_Timer.cancel();
            }
        CloseSockets();
        System.exit(0);
    }
        
    private static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    private int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }

    private String TruncateString(String sStr, int iMaxLength)
    {
        try
        {
            // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
            int iNullTerminatePosition = sStr.indexOf(0);
            if(-1 != iNullTerminatePosition)
            {
                sStr = sStr.substring(0, iNullTerminatePosition);
            }
            if(sStr.length()> iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.substring(0, iMaxLength);
            }
            return sStr;
        }
        catch (Exception Ex)
        {
            return "";
        }
    }
   
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        jTextArea_EventLog.append(sStr);       
    }

    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new DB_Arch_DBServer_DB1_Direct().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
        if(jButton_ClearLog == e.getSource())
        {
            jTextArea_EventLog.setText("");
        }
    }
    
    private JPanel jPanel_Info;
    private JLabel jLabel_Info1;
    private JLabel jLabel_Info2;
    private JLabel jLabel_Info3;
    private JPanel jPanel_Data;
    private JLabel jLabel_CustomerID;
    private JTextArea jTextArea_CustomerID;
    private JLabel jLabel_CustomerName;
    private JTextArea jTextArea_CustomerName;
    private JPanel jPanel_Controls;
    private JButton jButton_ClearLog;
    private JButton jButton_Done;
    private JPanel jPanel_EventLog;
    private JTextArea jTextArea_EventLog;
    private JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_Info1 = new JLabel();
        jLabel_Info1.setText("Note: This component forms part of a demonstration of aspects of connectivity,");
        jLabel_Info2 = new JLabel();
        jLabel_Info2.setText("coupling, architecture and transparency. As such it has only partial database");
        jLabel_Info3 = new JLabel();
        jLabel_Info3.setText("functionality; the data displayed is hardcoded.");
        jPanel_Info = new JPanel();
        jPanel_Info.setPreferredSize(new Dimension(480, 100));
        jPanel_Info.setBorder(BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Info);
        jPanel_Info.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jLabel_Info1)
                .add(jLabel_Info2)
                .add(jLabel_Info3));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel1Layout.createSequentialGroup()
                    .add(jLabel_Info1)
                    .add(jLabel_Info2)
                    .add(jLabel_Info3)));

        jLabel_CustomerID = new JLabel();
        jLabel_CustomerID.setText("Customer ID");
        jTextArea_CustomerID = new JTextArea();
        jTextArea_CustomerID.setMaximumSize(new Dimension(80, 20));
        jLabel_CustomerName = new JLabel();
        jLabel_CustomerName.setText("Customer Name");
        jTextArea_CustomerName = new JTextArea();
        jTextArea_CustomerName.setMaximumSize(new Dimension(120, 20));
        jPanel_Data = new JPanel();
        jPanel_Data.setPreferredSize(new Dimension(480, 200));
        jPanel_Data.setBorder(BorderFactory.createTitledBorder("Data held by DB Server"));

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_Data);
        jPanel_Data.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                    .add(jPanel2Layout.createSequentialGroup()
                        .add(jPanel2Layout.createParallelGroup()
                            .add(jLabel_CustomerID)
                            .add(jTextArea_CustomerID))
                        .addContainerGap(30, 30)
                        .add(jPanel2Layout.createParallelGroup()
                            .add(jLabel_CustomerName)
                            .add(jTextArea_CustomerName)))));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerID)
                    .addContainerGap(10, 10)
                    .add(jTextArea_CustomerID))
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerName)
                    .addContainerGap(10, 10)
                    .add(jTextArea_CustomerName)));            
        
        jButton_ClearLog = new javax.swing.JButton();
        jButton_ClearLog.setText("Clear Log entries");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new javax.swing.JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls = new javax.swing.JPanel();
        jPanel_Controls.setPreferredSize(new Dimension(250, 200));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jButton_ClearLog)
            .add(jButton_Done));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel3Layout.createSequentialGroup()
                .add(jButton_ClearLog)
                .addContainerGap(10, 10)
                .add(jButton_Done)));
        jPanel3Layout.linkSize(new java.awt.Component[] {jButton_ClearLog, jButton_Done});

        jTextArea_EventLog = new javax.swing.JTextArea();
        jScrollPane_EventLog = new javax.swing.JScrollPane(jTextArea_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog = new javax.swing.JPanel();
        jPanel_EventLog.setPreferredSize(new Dimension(330, 400));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event Log (Diagnostic)"));
        
        org.jdesktop.layout.GroupLayout jPanel4Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel4Layout);
        jPanel4Layout.setHorizontalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));
        jPanel4Layout.setVerticalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));        
         
        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(480, 400));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)            
                .add(jPanel_Info)
                .add(layout.createSequentialGroup()
                    .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)  
                        .add(jPanel_Data)
                        .add(jPanel_Controls))
                    .add(layout.createParallelGroup()  
                        .add(jPanel_EventLog)))));
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
                .add(jPanel_Info)
                .add(layout.createParallelGroup()  
                    .add(layout.createSequentialGroup()
                        .add(jPanel_Data)
                        .addContainerGap(100, 100)
                        .add(jPanel_Controls)
                        .addContainerGap(100, 100))
                    .add(layout.createSequentialGroup()
                        .add(jPanel_EventLog)))));

        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        setTitle("DB Arch DBServer (Type 1, Port 8002)");
        pack();
    }                                             
}