/*	File			MediaShare_Peer_Disconnect.java
	Purpose			Peer-peer demonstration
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			March 2015
*/
package MediaShare_Peer_Disconnect;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.*;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.*;
import java.util.*;
    
public class MediaShare_Peer_Disconnect extends javax.swing.JFrame implements ActionListener
{
    private static int PEER_DISCOVERY_PORT = 8004;      // Receive broadcast self-advertisement messages from other peers
    private static int PEER_APPLICATION_PORT = 8005;    // Receive point-to-point UDP messages from other peers
    private static int MAX_SONG_TITLE_LEN = 40;
    private static int MAX_ARTIST_NAME_LEN = 30;
    private static int MAX_DATA_ROWS_LOCAL = 4;
    private static int MAX_DATA_ROWS = MAX_DATA_ROWS_LOCAL * KnownPeers.MAX_KNOWN_PEERS;
    
    // Variables added to support detection of peer departure
    public static int HeartbeatClock_LIMIT = 1000;
    public static int PeerDeparted_Threshold_HeartbeatClockInterval = 3;

    public enum MessageType {PEER_SELF_ADVERTISEMENT, REQUEST_MEDIA_LIST, SEND_MEDIA_RECORD};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Application_Message_PDU
    {
        MessageType eMessageType;   // Offset 0
        int iSender_Random_ID;      // Offset 4
        Boolean bDataValid;         // Offset 8
        byte[] cSongTitle;          // Offset 9 - 49 inclusive
        byte[] cArtistName;         // Offset 50 - 80 inclusive
        // There is a 3-byte pad in the message as actually transmitted by the C++ version to truncate the message on a 4-byte boundary
    };

    private static int Application_Message_Size = 84;   // For serialisation / deserialisation (for compatibility with C++ version)

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class PeerSelfAdvertisement_Message_PDU
    {
        MessageType eMessageType;   // Offset 0
        int iSender_Random_ID;      // Offset 4
        // PeerAddressDetails in C++ 'SockAddr' format for compatibility;
        int iAddressFamily;         // Offset 8
        int iPort;                  // Offset 10-11
        byte iIP_Address_Byte_1;    // Offset 12
        byte iIP_Address_Byte_2;    // Offset 13
        byte iIP_Address_Byte_3;    // Offset 14
        byte iIP_Address_Byte_4;    // Offset 15
        // The C++ version includes a 16-byte Sockaddr_IN structure to hold the Address and port value (see serialisation / deserialisation code)
    };

    private static int PeerSelfAdvertisement_Message_PDU_Size = 24;   // For serialisation / deserialisation (for compatibility with C++ version)

    // Data Row and Table structures (hard coded data serves as mock of real database containing media files (songs))
    final class DataRow_LOCAL 
    {
        String sSongTitle;
        String sArtistName;
    };

    final class DataRow
    {
        String sSongTitle;
        String sArtistName;
        int iProviderPeerID;
    };

    DataRow_LOCAL[] m_LocalDataRowArray;
    DataRow[] m_DataRowArray; // Array of data row structures

    int m_iPeer_Random_ID;
    int m_iPeer_MediaConfiguration;
    int m_iHeartbeatClock;  // Used to measure time intervals between heartbeats from Peers

    MulticastSocket m_PeerDiscoverySock;
    DatagramSocket m_ApplicationSock;	
    InetAddress m_Local_Address;
    InetAddress m_Broadcast_IPAddress;

    Timer m_Timer_Receive;
    Timer m_Timer_Send_ServerAdvertisement;
        
    DefaultListModel m_listModel_Song; // For use with JList
    DefaultListModel m_listModel_Artist;
    DefaultListModel m_listModel_Provider;
    
    KnownPeers m_KnownPeers; // KnownPeers member variable (represents an object of the KnownPeers class created by the refactoring exercise)
                 
    public MediaShare_Peer_Disconnect()
    {
        m_KnownPeers = new KnownPeers(this);
        m_iHeartbeatClock = 0;	// Used to measure time intervals between heartbeats from Peers

        m_listModel_Song = new DefaultListModel();
        m_listModel_Artist = new DefaultListModel();
        m_listModel_Provider = new DefaultListModel();
        initComponents();
        Random rnd = new Random();
        m_iPeer_Random_ID = rnd.nextInt(10000);
        m_iPeer_MediaConfiguration = rnd.nextInt(10); // Randomly select a media configuration betwen 0 and 9
        InitialiseDataTable_LOCAL();
        InitialiseDataTable_FULL();
        DisplayDataTable();
        GetIPAddresses();
        Initialise_PeerDiscoverySock();
        Initialise_ApplicationSock();
        InitialiseControls();
        EnablePeer();
    }

    private void InitialiseControls()    
    {    
        jTextField_RandomConfigID.setEnabled(false);
        jTextField_RandomConfigID.setText(Integer.toString(m_iPeer_MediaConfiguration));
        jTextField_PeerID.setEnabled(false);
        jTextField_PeerID.setText(Integer.toString(m_iPeer_Random_ID));
        jList_Song.setEnabled(false);
        jList_Song.setBackground(Color.white);
        jList_Song.setSelectionBackground(Color.white);
        JScrollPane_Song.setEnabled(false);
        jList_Artist.setEnabled(false);
        jList_Artist.setBackground(Color.white);
        jList_Artist.setSelectionBackground(Color.white);
        JScrollPane_Artist.setEnabled(false);
        jList_Provider.setEnabled(false);
        jList_Provider.setBackground(Color.white);
        jList_Provider.setSelectionBackground(Color.white);
        JScrollPane_Provider.setEnabled(false);
        jButton_ClearLog.setEnabled(true); 
        jButton_Done.setEnabled(true); 
        jTextArea_EventLog.setEnabled(false);
    }

    void GetIPAddresses()
    {   
        String sHostAddress = "";
        try
        {
            m_Local_Address = InetAddress.getLocalHost();
            sHostAddress = m_Local_Address.toString(); // Expected format: "Hostname/192.168.100.5"
            String sAddressDisplay = "Local address " + sHostAddress;
            WriteEventLogEntry(sAddressDisplay);            
            String sStr = String.format("Send / Recv application messages port %d",PEER_APPLICATION_PORT);
            WriteEventLogEntry(sStr);
            sStr = String.format("B'cast / Recv peer discovery messages port %d",PEER_DISCOVERY_PORT);
            WriteEventLogEntry(sStr);
            m_Broadcast_IPAddress = GetLocalBroadcastAddress_From_LocalHostAddress(m_Local_Address);
            byte[] bAddress = m_Broadcast_IPAddress.getAddress();
            sStr = String.format("B'cast address for peer discovery %s", DottedIPAddressString_from_4ByteArray(bAddress));
            WriteEventLogEntry(sStr);
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"MediaShare_Peer",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
    }
    
    void InitialiseDataTable_LOCAL()	// Populate  m_DataTable_LOCAL (hardcoded data)
    {
        m_LocalDataRowArray = new DataRow_LOCAL[MAX_DATA_ROWS_LOCAL];

        for (int iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
        {   // Initialise m_DataTable_LOCAL (all strings blank)
            m_LocalDataRowArray[iIndex] = new DataRow_LOCAL();
            m_LocalDataRowArray[iIndex].sSongTitle = "";
            m_LocalDataRowArray[iIndex].sArtistName = "";
        }

        switch(m_iPeer_MediaConfiguration) // Values 0 - 9 chosen randomly to simulate different peer instances
        {
            case 0:
                m_LocalDataRowArray[0].sSongTitle = "Bicycle Race";
                m_LocalDataRowArray[0].sArtistName = "Queen";
                m_LocalDataRowArray[1].sSongTitle = "Sir Duke";
                m_LocalDataRowArray[1].sArtistName = "Stevie Wonder";
                break;
            case 1:
                m_LocalDataRowArray[0].sSongTitle = "Undercover Of The Night";
                m_LocalDataRowArray[0].sArtistName = "The Rolling Stones";
                m_LocalDataRowArray[1].sSongTitle = "Dancing In The Moonlight";
                m_LocalDataRowArray[1].sArtistName = "Thin Lizzy";
                m_LocalDataRowArray[2].sSongTitle = "Get Back";
                m_LocalDataRowArray[2].sArtistName = "The Beatles";
                break;
            case 2:
                m_LocalDataRowArray[0].sSongTitle = "Living For The City";
                m_LocalDataRowArray[0].sArtistName = "Stevie Wonder";
                break;
            case 3:
                m_LocalDataRowArray[0].sSongTitle = "Dont Stop Me Now";
                m_LocalDataRowArray[0].sArtistName = "Queen";
                break;
            case 4:
                m_LocalDataRowArray[0].sSongTitle = "Start Me UP";
                m_LocalDataRowArray[0].sArtistName = "The Rolling Stones";
                m_LocalDataRowArray[1].sSongTitle = "My Generation";
                m_LocalDataRowArray[1].sArtistName = "The Who";
                break;
            case 5:
                m_LocalDataRowArray[0].sSongTitle = "Let It Be";
                m_LocalDataRowArray[0].sArtistName = "The Beatles";
                m_LocalDataRowArray[1].sSongTitle = "Rock And Roll";
                m_LocalDataRowArray[1].sArtistName = "Led Zeppelin";
                m_LocalDataRowArray[2].sSongTitle = "Master Blaster";
                m_LocalDataRowArray[2].sArtistName = "Stevie Wonder";
                m_LocalDataRowArray[3].sSongTitle = "Leaving Here";
                m_LocalDataRowArray[3].sArtistName = "The Who";
                break;
            case 6:
                m_LocalDataRowArray[0].sSongTitle = "House Of The Rising Sun";
                m_LocalDataRowArray[0].sArtistName = "The Animals";
                m_LocalDataRowArray[1].sSongTitle = "The Battle Of Evermore";
                m_LocalDataRowArray[1].sArtistName = "Led Zeppelin";
                m_LocalDataRowArray[2].sSongTitle = "Sarah";
                m_LocalDataRowArray[2].sArtistName = "Thin Lizzy";
                break;
            case 7:
                m_LocalDataRowArray[0].sSongTitle = "Whole Lotta Love";
                m_LocalDataRowArray[0].sArtistName = "Led Zeppelin";
                m_LocalDataRowArray[1].sSongTitle = "We've Gotta Get Out Of This Place";
                m_LocalDataRowArray[1].sArtistName = "The Animals";
                m_LocalDataRowArray[2].sSongTitle = "Seven Seas Of Rhye";
                m_LocalDataRowArray[2].sArtistName = "Queen";
                m_LocalDataRowArray[3].sSongTitle = "Ticket To Ride";
                m_LocalDataRowArray[3].sArtistName = "The Beatles";
                break;
            case 8:
                m_LocalDataRowArray[0].sSongTitle = "Hey Jude";
                m_LocalDataRowArray[0].sArtistName = "The Beatles";
                break;
            case 9:
                m_LocalDataRowArray[0].sSongTitle = "Waiting For An Alibi";
                m_LocalDataRowArray[0].sArtistName = "Thin Lizzy";
                m_LocalDataRowArray[1].sSongTitle = "Harlem Shuffle";
                m_LocalDataRowArray[1].sArtistName = "The Rolling Stones";
                break;
        }
    }

    void InitialiseDataTable_FULL()	// Upate DataTable
    {
        m_DataRowArray = new DataRow[MAX_DATA_ROWS];

        int iIndex;
        for(iIndex = 0; iIndex < MAX_DATA_ROWS; iIndex++)
        {   // Initialise DataTable (all strings blank)
            m_DataRowArray[iIndex] = new DataRow();
            m_DataRowArray[iIndex].sSongTitle = "";
            m_DataRowArray[iIndex].sArtistName = "";
            m_DataRowArray[iIndex].iProviderPeerID = -1;
        }

        // Copy local data rows into full data table, until there are no more valid entries, or the maximum number of local entries is reached
        for(iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
        {
            if (false == m_LocalDataRowArray[iIndex].sSongTitle.equals(""))
            {
                m_DataRowArray[iIndex].sSongTitle = m_LocalDataRowArray[iIndex].sSongTitle;
                m_DataRowArray[iIndex].sArtistName = m_LocalDataRowArray[iIndex].sArtistName;
                m_DataRowArray[iIndex].iProviderPeerID = m_iPeer_Random_ID;
            }
        }
    }

    void Add_Row_To_DataTable_FULL(String sSongTitle, String sArtistName, int iProviderPeerID)	// Upate DataTable
    {
        int iIndex;
        iIndex = 0;
        while (false == m_DataRowArray[iIndex].sSongTitle.equals(""))
        {   // Skip populated entries
            iIndex++;
        }
        // iIndex now points to next available entry
        if(iIndex < MAX_DATA_ROWS)
        {
            m_DataRowArray[iIndex].sSongTitle = sSongTitle;
            m_DataRowArray[iIndex].sArtistName = sArtistName;
            m_DataRowArray[iIndex].iProviderPeerID = iProviderPeerID;
        }
        DisplayDataTable();
    }

    void DisplayDataTable()
    {	
        m_listModel_Song.clear();
        m_listModel_Artist.clear();
        m_listModel_Provider.clear();
        for(int iIndex = 0; iIndex < MAX_DATA_ROWS; iIndex++)
        {   // Display all valid rows
            if (false == m_DataRowArray[iIndex].sSongTitle.equals(""))
            {
                m_listModel_Song.addElement(m_DataRowArray[iIndex].sSongTitle);
                m_listModel_Artist.addElement(m_DataRowArray[iIndex].sArtistName);
                m_listModel_Provider.addElement(Integer.toString(m_DataRowArray[iIndex].iProviderPeerID));
            }
        }
    }

    void Initialise_PeerDiscoverySock()
    {	
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_PeerDiscoverySock = new MulticastSocket(PEER_DISCOVERY_PORT); // create socket and bind to the port for receiveing
            m_PeerDiscoverySock.setBroadcast(true);     // Enable broadcast sends
            m_PeerDiscoverySock.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"MediaShare Peer",0);
           CloseAndExit();
        }
    }

    void Initialise_ApplicationSock()
    {		
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_ApplicationSock = new DatagramSocket(PEER_APPLICATION_PORT); // create socket and bind to the port for receiveing
            m_ApplicationSock.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"MediaShare Peer",0);
           CloseAndExit();
        }
    }

    void EnablePeer()  // Invoked on startup
    {
        Start_Timer_Receive();   // As all activity is non-blocking
        Start_Timer_Send_ServerAdvertisement();
    }

    void Start_Timer_Receive()
    {
        m_Timer_Receive = new Timer();
        m_Timer_Receive.scheduleAtFixedRate(new OnTimedEvent_Receive(), 500, 500);
    }

    private class OnTimedEvent_Receive extends TimerTask
    {
        public void run()
        {
            DoReceive_ApplicationMessages();    // Check for application messages from known peers
            DoReceive_PeerAdvertisements();     // Check for self advertisement peers and discover previously unknown peers
        }
    }

    void Start_Timer_Send_ServerAdvertisement()
    {
        m_Timer_Send_ServerAdvertisement = new Timer();
        m_Timer_Send_ServerAdvertisement.scheduleAtFixedRate(
            new OnTimedEvent_ServerAdvertisement(), 5000, 5000);// 5 second interval between sending ServerAdvertisement messages
    }

    private class OnTimedEvent_ServerAdvertisement extends TimerTask
    {
        public void run()
        {
            Send_SelfAdvertisement();  // Broadcast Self Advertisement message to other Peers
            Send_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeers(); 

            m_iHeartbeatClock++;    // Used to measure time intervals between heartbeats from Peers
            m_iHeartbeatClock %= HeartbeatClock_LIMIT;	// m_iHeartbeatClock counts from 0 to 999, then is reset to 0 and start counting again

            // Check all known-peers for disconnects (we have not detected heartbeats from them for three heartbeat-send intervals)
            m_KnownPeers.CheckPeerHeartbeatStatus(m_iHeartbeatClock);
        }
    }
    
    void Send_SelfAdvertisement()
    {
        PeerSelfAdvertisement_Message_PDU SelfAdvertisement_Message = new PeerSelfAdvertisement_Message_PDU();
        SelfAdvertisement_Message.eMessageType = MessageType.PEER_SELF_ADVERTISEMENT;
        SelfAdvertisement_Message.iSender_Random_ID = m_iPeer_Random_ID;
        byte[] bAddress = m_Local_Address.getAddress();
        SelfAdvertisement_Message.iIP_Address_Byte_1 = bAddress[0];
        SelfAdvertisement_Message.iIP_Address_Byte_2 = bAddress[1];
        SelfAdvertisement_Message.iIP_Address_Byte_3 = bAddress[2];
        SelfAdvertisement_Message.iIP_Address_Byte_4 = bAddress[3];
        SelfAdvertisement_Message.iPort = PEER_APPLICATION_PORT;
        SelfAdvertisement_Message.iAddressFamily = 2;   // For C++ compatibility, value 2 = AF_INET
        byte[] bSendBuf = Serialize_PeerSelfAdvertisement_Message_PDU(SelfAdvertisement_Message);

        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                                            m_Broadcast_IPAddress, PEER_DISCOVERY_PORT); 
            m_PeerDiscoverySock.send(SendPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return;
        }
    }

    void SendApplicationMessage(Application_Message_PDU Message, byte[] bAddress)
    {
        // The sendto addresss is set dynamically (depending on which known Peer the application message is to be sent to)
        byte[] bSendBuf = Serialize_Application_Message_PDU(Message);
        try
        {
            InetAddress send_IPAddress = InetAddress.getByAddress(bAddress);
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                                            send_IPAddress, PEER_APPLICATION_PORT); 
            m_ApplicationSock.send(SendPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return;
        }
    }

    void Send_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeers()
    {   // A race can occur if an application message (especially the first one sent: REQUEST_MEDIA_LIST) is sent
        // immediately after a remote peer is discovered. This is because discovery is two-sided and the fact that
        // the local peer has discovered the remote peer does not guarantee that the remote peer has discovered the 
        // local peer. Received application messages that do not originate from known peers (i.e. in local m_Known_Peers_List)
        // are ignored, hence the need for a special means of issueing REQUEST_MEDIA_LIST messages (which has the effect of 
        // decoupling peer discovery and requesting Media data):
        // - For each discovered peer an entry is made in the m_Known_Peers_List array. 
        // - For each entry the bMediaDataReceived flag is initially false, only set true once the Media data has been received.
        // - Each time this method is invoked (via the timer) the m_Known_Peers_List is scanned to see if there are any new
        // peers for which a REQUEST_MEDIA_LIST must be issued.
        KnownPeers.Known_Peer_details KPD;
        for(int iIndex = 0; iIndex < KnownPeers.MAX_KNOWN_PEERS; iIndex++)
            {
            KPD = m_KnownPeers.get_Known_Peers_List_entry_byIndex(iIndex);
            if (true == KPD.bInUse)
                    {
                if (false == KPD.bMediaDataReceived)
                {  // Do not yet have Media data; request it now
                    Application_Message_PDU Message = new Application_Message_PDU();
                    Message.eMessageType = MessageType.REQUEST_MEDIA_LIST;
                    Message.bDataValid = false;
                    Message.iSender_Random_ID = m_iPeer_Random_ID;
                    SendApplicationMessage(Message, KPD.bAddress);
                    String sStr = String.format("REQUEST_MEDIA_LIST request sent to %s", 
                            DottedIPAddressString_from_4ByteArray(KPD.bAddress));
                    WriteEventLogEntry(sStr);
                }
            }
        }
    }

    void Send_MediaFilesDetails_To_requestor(byte[] bAddress)
    {
        Application_Message_PDU Message = new Application_Message_PDU();
        Message.eMessageType = MessageType.SEND_MEDIA_RECORD;
        Message.iSender_Random_ID = m_iPeer_Random_ID;
        Message.bDataValid = true;
        Message.cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
        Message.cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];

        for(int iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
        {   // Send all valid rows, one by one, (determined by the song title field not being the empty string "")
            // Send only the data concerning media files held locally
            if(false == m_LocalDataRowArray[iIndex].sSongTitle.equals(""))
            {
                Message.cSongTitle = m_LocalDataRowArray[iIndex].sSongTitle.getBytes();
                Message.cArtistName = m_LocalDataRowArray[iIndex].sArtistName.getBytes();
                String sStr = String.format("SEND_MEDIA_RECORD %s,%s to %s", 
                                        m_LocalDataRowArray[iIndex].sSongTitle,
                                        m_LocalDataRowArray[iIndex].sArtistName, 
                                        DottedIPAddressString_from_4ByteArray(bAddress));
                WriteEventLogEntry(sStr);
                SendApplicationMessage(Message, bAddress);
            }
        }			
        Message.cSongTitle = (new String("--END--")).getBytes(); // Signals end of data
        SendApplicationMessage(Message, bAddress);
    }

    void DoReceive_ApplicationMessages()
    {   // Check for Application messages (from ANY other peers)
        byte[] bRecvBuf = new byte[Application_Message_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, Application_Message_Size);
        InetAddress SenderIPAddress;

        try
        {
            m_ApplicationSock.receive(RecvPacket);
            SenderIPAddress = RecvPacket.getAddress(); // Place the sender's address into the m_SenderIPAddress
        }
        catch(SocketTimeoutException ex)
        {
             return;
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            return;
        }
        if (Application_Message_Size == RecvPacket.getLength())
        {
            Application_Message_PDU Message = DeSerialize_ApplicationMessage(bRecvBuf);
            int iIndex = m_KnownPeers.GetIndexOf_KnownPeer_InPeersList(Message.iSender_Random_ID);
            if(-1 == iIndex)
            {
                WriteEventLogEntry("Sender ID not found in Known Peers array");
                return;
            }

            // If here; found address of sender to display, and for replying to, in Known Peers array
            KnownPeers.Known_Peer_details KPD = m_KnownPeers.get_Known_Peers_List_entry_byIndex(iIndex);
            
            byte[] bAddress = SenderIPAddress.getAddress();
            String sStr = "";
            switch(Message.eMessageType)
            {
                case REQUEST_MEDIA_LIST:
                    sStr = String.format("REQUEST_MEDIA_LIST from %d %s", 
                            Message.iSender_Random_ID, DottedIPAddressString_from_4ByteArray(bAddress));
                    WriteEventLogEntry(sStr);
                    Send_MediaFilesDetails_To_requestor(KPD.bAddress);
                    break;
                case SEND_MEDIA_RECORD:
                    sStr = new String(Message.cSongTitle);
                    if(true == sStr.contains("--END--"))
                    {   // Signals that all media data has been sent by the remote peer
                        m_KnownPeers.set_Known_Peers_List_entry_bMediaDataReceived_flag_byIndex(iIndex);
                        // Prevents further REQUEST_MEDIA_LIST messages being sent to the particular remote peer
                    }
                    else
                    {	// Update the data table and display
                        if (false == KPD.bMediaDataReceived)
                        {
                            sStr = String.format("SEND_MEDIA_RECORD from %d %s", 
                                    Message.iSender_Random_ID, DottedIPAddressString_from_4ByteArray(bAddress));
                            WriteEventLogEntry(sStr);
                            sStr = new String(Message.cSongTitle);
                            String sSong = TruncateString(sStr, sStr.length());
                            sStr = new String(Message.cArtistName);
                            String sArtist = TruncateString(sStr, sStr.length());
                            sStr = String.format("MEDIA received, Song:%s  Artist:%s", sSong, sArtist);
                            WriteEventLogEntry(sStr);
                            Add_Row_To_DataTable_FULL(sSong, sArtist, Message.iSender_Random_ID);
                        }
                    }
                    break;
            }
        }
    }

    void DoReceive_PeerAdvertisements()
    {		
        byte[] bRecvBuf = new byte[PeerSelfAdvertisement_Message_PDU_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, PeerSelfAdvertisement_Message_PDU_Size);
        try
        {
            m_PeerDiscoverySock.receive(RecvPacket);
        }
        catch(SocketTimeoutException ex)
        {
             return;
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            return;
        }

        if (PeerSelfAdvertisement_Message_PDU_Size == RecvPacket.getLength())
        {
            PeerSelfAdvertisement_Message_PDU PSA_Message = DeSerialize_PeerSelfAdvertisementMessage(bRecvBuf);
            byte[] bSenderAddress = new byte[4]; // Represent the sender's address in the form of a byte array
            bSenderAddress[0] = PSA_Message.iIP_Address_Byte_1;
            bSenderAddress[1] = PSA_Message.iIP_Address_Byte_2;
            bSenderAddress[2] = PSA_Message.iIP_Address_Byte_3;
            bSenderAddress[3] = PSA_Message.iIP_Address_Byte_4;
            
            switch (PSA_Message.eMessageType)
            {
                case PEER_SELF_ADVERTISEMENT:
                byte[] bLocalAddress = m_Local_Address.getAddress();
                if (false == Arrays.equals(bLocalAddress,bSenderAddress))
                {   // Check if message is from self - if so ignore
                    if(false == m_KnownPeers.CheckIfPeerExistsIn_KnownPeersList(PSA_Message.iSender_Random_ID,bSenderAddress))
                    {   // Check if peer is already known (need combination of ID and address), Here if not yet known
                        if(true == m_KnownPeers.Add_NewlyDiscoveredPeerTo_KnownPeersList(PSA_Message.iSender_Random_ID,bSenderAddress))
                        {
                            String sStr = String.format("New peer discovered, ID:%d  Address %s", 
                                    PSA_Message.iSender_Random_ID, DottedIPAddressString_from_4ByteArray(bSenderAddress));
                            WriteEventLogEntry(sStr);
                        }
                        else
                        {
                            String sStr = String.format("New peer discovered ID:%d !No space in KnownPeers array!",
                                                                                    PSA_Message.iSender_Random_ID);
                            WriteEventLogEntry(sStr);							 
                        }
                    }
                    else
                    {
                        // Received PEER_SELF_ADVERTISEMENT message from an already-known peer - Use this as a heartbeat message
                        // So that the local peer can detect when one of its known peers departs (and thus its resources are no longer available)
                        int iIndex = m_KnownPeers.GetIndexOf_KnownPeer_InPeersList(PSA_Message.iSender_Random_ID);
                        if(-1 == iIndex)
                        {
                            return;
                        }
                        // If here; found the index position of the sender of the heartbeat, in the Known_Peer_details array
                        m_KnownPeers.set_Known_Peers_HeartbeatStatus_byIndex(iIndex, m_iHeartbeatClock);
                    }
                }
                break;
            }
        }
    }

public void Callback_PeerDeparted_UpdateStatus(int iPeerID)
    {
        String sStr = String.format("Peer disconnected ID:%d - removing published resources",iPeerID);
        WriteEventLogEntry(sStr);							 
        for(int iIndex = 0; iIndex < MAX_DATA_ROWS; iIndex++)
        {   // Search for all rows 'provided' by the departed peer
            if (m_DataRowArray[iIndex].iProviderPeerID == iPeerID)
            {	// Clear entries previously provided by the departed peer
                m_DataRowArray[iIndex] = new DataRow();
                m_DataRowArray[iIndex].sSongTitle = "";
                m_DataRowArray[iIndex].sArtistName = "";
                m_DataRowArray[iIndex].iProviderPeerID = -1;
            }
        }
        DisplayDataTable(); // Show updated table after removal of no-longer-available resources
    }

    public int GetHeartbeatClock()
    {
        return m_iHeartbeatClock;
    }
     
    PeerSelfAdvertisement_Message_PDU DeSerialize_PeerSelfAdvertisementMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
        PeerSelfAdvertisement_Message_PDU result_Message = new PeerSelfAdvertisement_Message_PDU();
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);
        // Deserialisation of port number
        // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
        // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
        // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
        // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
        result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
        result_Message.iPort = (int)((256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/);
        result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
        result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
        result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
        result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
        return result_Message;
    }    
    
    Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
        Application_Message_PDU result_Message = new Application_Message_PDU();
        //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);
        result_Message.bDataValid = true;
        if (0 == bRecdByteArray[8])
        {
            result_Message.bDataValid = false;
        }
        result_Message.cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
        System.arraycopy(bRecdByteArray/*src*/, 9, result_Message.cSongTitle/*dest*/, 0, MAX_SONG_TITLE_LEN/*count*/);
        result_Message.cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];
        System.arraycopy(bRecdByteArray/*src*/, 50, result_Message.cArtistName/*dest*/, 0, MAX_ARTIST_NAME_LEN/*count*/);
        return result_Message;
    }

    byte[] Serialize_PeerSelfAdvertisement_Message_PDU(PeerSelfAdvertisement_Message_PDU Message)
    {   // Convert an ServerAdvertisement_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[PeerSelfAdvertisement_Message_PDU_Size];        // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bByteArray[8] = (byte)Message.iAddressFamily;   // For C++ compatibility, value 2 = AF_INET
        // Serialisation of port number
        // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
        // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
        // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
        // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
        bByteArray[10] = (byte)(Message.iPort / 256);
        bByteArray[11] = (byte)(Message.iPort % 256);
        bByteArray[12] = Message.iIP_Address_Byte_1;
        bByteArray[13] = Message.iIP_Address_Byte_2;
        bByteArray[14] = Message.iIP_Address_Byte_3;
        bByteArray[15] = Message.iIP_Address_Byte_4;
        return bByteArray;
    }

    byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
    {   // Convert an Application_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bByteArray[8] = 0; // false (bDataValid)
        if(true == Message.bDataValid)
        {
            bByteArray[8] = 1; // true (bDataValid)
        }
        if (null != Message.cSongTitle)
        {
            System.arraycopy(Message.cSongTitle/*src*/, 0, bByteArray/*dest*/, 9, Message.cSongTitle.length/*count*/);
        }
        if (null != Message.cArtistName)
        {
            System.arraycopy(Message.cArtistName/*src*/, 0, bByteArray/*dest*/, 50, Message.cArtistName.length/*count*/);
        }
        return bByteArray;
    }
   
    InetAddress GetLocalBroadcastAddress_From_LocalHostAddress(InetAddress LocalAddress) throws Exception
    {   // Assumes IPv4 address
        try
        {
            byte[] bLocalBroadcast = LocalAddress.getAddress(); // Initialise byte array with local host's address (e.g. 192.168.100.6)
            // Set the host portion of the address to the broadcast value, depending on the address class
            // Class A from 0.0.0.0 to 127.255.255.255   will be converted to the broadcast address x.255.255.255 
            // Class B from 128.0.0.0 to 191.255.255.255   will be converted to the broadcast address x.x.255.255  
            // Class C 192.0.0.0 to 223.255.255.255   will be converted to the broadcast address x.x.x.255 
            // Class D and E (224.0.0.0 to 225.255.255.254)   will be converted to the broadcast address 255.255.255.255 
            
            // For comparison purposes, the first byte of the address is converted to an int value (to ensure operations are performed 'unsigned')
            int iMostSignificantAddrsssByte = (int) bLocalBroadcast[0] & 0xFF;
            if(iMostSignificantAddrsssByte >= 0 && iMostSignificantAddrsssByte <= 127)
            {   // Class A address
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 128 && iMostSignificantAddrsssByte <= 191)
            {   // Class B address
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 192 && iMostSignificantAddrsssByte <= 223)
            {   // Class C address
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 224)
            {   // Class D / E address
                bLocalBroadcast[0] = (byte)255;
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            InetAddress LocalBroadcastAddress = InetAddress.getByAddress(bLocalBroadcast);
            return LocalBroadcastAddress;
        }
        catch (Exception Ex)
        {
            throw(Ex);
        }
    }

    String DottedIPAddressString_from_4ByteArray(byte[] bAddress)
    {
        String sStr = "";
        int iComponent = 0;
        if(4 != bAddress.length)
        {
            return sStr;    // If byte array is not 4 bytes return empty string
        }
        iComponent = bAddress[0] & 0xFF;
        sStr = Integer.toString(iComponent) + ".";
        iComponent = bAddress[1] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[2] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[3] & 0xFF;            
        sStr += Integer.toString(iComponent);
        return sStr;
    }
    
    void CloseSockets()
    {
        try
        {
            m_PeerDiscoverySock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_ApplicationSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
        if(null != m_Timer_Receive)
        {
            m_Timer_Receive.cancel();
        }
        if(null != m_Timer_Send_ServerAdvertisement)
        {
            m_Timer_Send_ServerAdvertisement.cancel();
        }
        CloseSockets();
        System.exit(0);
    }
        
    private static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    private int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }

    private String TruncateString(String sStr, int iMaxLength)
    {
        try
        {
            // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
            int iNullTerminatePosition = sStr.indexOf(0);
            if(-1 != iNullTerminatePosition)
            {
                sStr = sStr.substring(0, iNullTerminatePosition);
            }
            if(sStr.length()> iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.substring(0, iMaxLength);
            }
            return sStr;
        }
        catch (Exception Ex)
        {
            return "";
        }
    }
   
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        jTextArea_EventLog.append(sStr);       
    }

    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new MediaShare_Peer_Disconnect().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
        if(jButton_ClearLog == e.getSource())
        {
            jTextArea_EventLog.setText("");
        }
    }
    
    private JPanel jPanel_Info;
    private JLabel jLabel_Info1;
    private JLabel jLabel_Info2;
    private JLabel jLabel_Info3;
    private JPanel jPanel_State;
    private JLabel jLabel_RandomConfigID1;
    private JLabel jLabel_RandomConfigID2;
    private JTextField jTextField_RandomConfigID;
    private JLabel jLabel_PeerID;
    private JTextField jTextField_PeerID;
    private JPanel jPanel_Data;
    private JLabel jLabel_Song1;
    private JLabel jLabel_Song2;
    private JList jList_Song;
    private JScrollPane JScrollPane_Song;
    private JLabel jLabel_Artist1;
    private JLabel jLabel_Artist2;
    private JList jList_Artist;
    private JScrollPane JScrollPane_Artist;
    private JLabel jLabel_Provider1;
    private JLabel jLabel_Provider2;
    private JList jList_Provider;
    private JScrollPane JScrollPane_Provider;
    private JPanel jPanel_Controls;
    private JButton jButton_ClearLog;
    private JButton jButton_Done;
    private JPanel jPanel_EventLog;
    private JTextArea jTextArea_EventLog;
    private JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_Info1 = new JLabel();
        jLabel_Info1.setText("Media data is hardcoded.");
        jLabel_Info2 = new JLabel();
        jLabel_Info2.setText("Ten different pre-determined configurations are provided to allow simulation of a realistic");
        jLabel_Info3 = new JLabel();
        jLabel_Info3.setText("peer-peer scenario in which peers have different resources that they share with other peers.");
        jPanel_Info = new JPanel();
        jPanel_Info.setPreferredSize(new Dimension(480, 100));
        jPanel_Info.setBorder(BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Info);
        jPanel_Info.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jLabel_Info1)
                .add(jLabel_Info2)
                .add(jLabel_Info3));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel1Layout.createSequentialGroup()
                    .add(jLabel_Info1)
                    .add(jLabel_Info2)
                    .add(jLabel_Info3)));

        jLabel_RandomConfigID1 = new JLabel();
        jLabel_RandomConfigID1.setText("Random media configuration");
        jLabel_RandomConfigID2 = new JLabel();
        jLabel_RandomConfigID2.setText("(represents different peer instances, 0-9)");
        jTextField_RandomConfigID = new JTextField();
        jTextField_RandomConfigID.setMaximumSize(new Dimension(40, 4));
        jLabel_PeerID = new JLabel();
        jLabel_PeerID.setText("Peer ID (diagnostic)");
        jTextField_PeerID = new JTextField();
        jTextField_PeerID.setMaximumSize(new Dimension(40, 4));
        jPanel_State = new JPanel();
        jPanel_State.setPreferredSize(new Dimension(380, 200));
        jPanel_State.setBorder(BorderFactory.createTitledBorder(""));
        
        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_State);
        jPanel_State.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                    .add(jPanel2Layout.createSequentialGroup()
                        .add(jPanel2Layout.createParallelGroup()
                            .add(jLabel_RandomConfigID1)
                            .add(jLabel_RandomConfigID2))                    
                            .addContainerGap(10, 10)
                            .add(jTextField_RandomConfigID))
                        .add(jPanel2Layout.createSequentialGroup()
                            .add(jLabel_PeerID)
                            .addContainerGap(125, 125)
                            .add(jTextField_PeerID)));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jPanel2Layout.createParallelGroup()
                       .add(jPanel2Layout.createSequentialGroup()
                           .add(jLabel_RandomConfigID1)
                           .add(jLabel_RandomConfigID2)
                           .addContainerGap(30, 30))
                       .add(jTextField_RandomConfigID))
                    .add(jPanel2Layout.createParallelGroup()
                       .add(jLabel_PeerID)
                       .add(jTextField_PeerID))));
        
        jLabel_Song1 = new JLabel();
        jLabel_Song1.setText("Song");
        jLabel_Song2 = new JLabel();
        jLabel_Song2.setText(" ");
        jList_Song = new JList(m_listModel_Song);
        jList_Song.setMaximumSize(new Dimension(310, 350));
        JScrollPane_Song = new javax.swing.JScrollPane(jList_Song,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jLabel_Artist1 = new JLabel();
        jLabel_Artist1.setText("Artist");
        jLabel_Artist2 = new JLabel();
        jLabel_Artist2.setText(" ");
        jList_Artist = new JList(m_listModel_Artist);
        jList_Artist.setMaximumSize(new Dimension(310, 350));
        JScrollPane_Artist = new javax.swing.JScrollPane(jList_Artist,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jLabel_Provider1 = new JLabel();
        jLabel_Provider1.setText("Provider");
        jLabel_Provider2 = new JLabel();
        jLabel_Provider2.setText("(Peer ID)");
        jList_Provider = new JList(m_listModel_Provider);
        jList_Provider.setMaximumSize(new Dimension(40, 350));
        JScrollPane_Provider = new javax.swing.JScrollPane(jList_Provider,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_Data = new JPanel();
        jPanel_Data.setPreferredSize(new Dimension(380, 350));
        jPanel_Data.setBorder(BorderFactory.createTitledBorder("Media files available"));

        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_Data);
        jPanel_Data.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                    .add(jPanel3Layout.createSequentialGroup()
                        .add(jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                            .add(jLabel_Song1)
                            .add(jLabel_Song2)
                            .add(JScrollPane_Song))
                        .add(jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                            .add(jLabel_Artist1)
                            .add(jLabel_Artist2)
                            .add(JScrollPane_Artist))
                        .add(jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                            .add(jLabel_Provider1)
                            .add(jLabel_Provider2)
                            .add(JScrollPane_Provider)))));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_Song1)
                    .add(jLabel_Song2)
                    .add(JScrollPane_Song))
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_Artist1)
                    .add(jLabel_Artist2)
                    .add(JScrollPane_Artist))            
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_Provider1)
                    .add(jLabel_Provider2)
                    .add(JScrollPane_Provider)));            
        
        jButton_ClearLog = new javax.swing.JButton();
        jButton_ClearLog.setText("Clear Log entries");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new javax.swing.JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls = new javax.swing.JPanel();
        jPanel_Controls.setPreferredSize(new Dimension(400, 100));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel4Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel4Layout);
        jPanel4Layout.setHorizontalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel4Layout.createSequentialGroup()
                    .addContainerGap(45, 45)
                    .add(jButton_Done)
                    .addContainerGap(80, 80)
                    .add(jButton_ClearLog)
                    .addContainerGap(45, 45)));
        jPanel4Layout.setVerticalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jButton_Done)
                .add(jButton_ClearLog));

        jTextArea_EventLog = new javax.swing.JTextArea();
        jTextArea_EventLog.setMaximumSize(new Dimension(450, 500));
        jScrollPane_EventLog = new javax.swing.JScrollPane(jTextArea_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog = new javax.swing.JPanel();
        jPanel_EventLog.setPreferredSize(new Dimension(450, 500));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event Log (Diagnostic)"));
        
        org.jdesktop.layout.GroupLayout jPanel5Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel5Layout);
        jPanel5Layout.setHorizontalGroup(
            jPanel5Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));
        jPanel5Layout.setVerticalGroup(
            jPanel5Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));        
         
        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(800, 500));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)            
                .add(jPanel_Info)
                .add(layout.createSequentialGroup()
                    .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)  
                        .add(jPanel_State)
                        .add(jPanel_Data)
                        .add(jPanel_Controls))
                    .add(layout.createParallelGroup()  
                        .add(jPanel_EventLog)))));
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
                .add(jPanel_Info)
                .add(layout.createParallelGroup()  
                    .add(layout.createSequentialGroup()
                        .add(jPanel_State)
                        .addContainerGap(20, 20)
                        .add(jPanel_Data)
                        .addContainerGap(20, 20)
                        .add(jPanel_Controls))
                    .add(layout.createSequentialGroup()
                        .add(jPanel_EventLog)))));

        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        setTitle("Peer - Peer demonstration (Media sharing example)");
        pack();
    }                                             
}