/*	File			Replication_Client.java
	Purpose			Replication demonstration
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			March 2015
*/
package Replication_Client;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.*;
    
public class Replication_Client extends javax.swing.JFrame implements ActionListener
{
    private static int REPLICATION_EXTERNAL_PORT = 8006;  // Client to Server
    private static int REPLICATION_UPDATE_PORT = 8007;	// Server to Server - Update propagation
    private static int REPLICATION_HEARTBEAT_PORT = 8008; // Server to Server - Master heartbeat broadcasts
    private static int REPLICATION_SERVER_ADVERTISEMENT_PORT = 8009; // Server to client
    private static int MAX_CUST_NAME_LEN = 20;

    public enum ServerState {MASTER, BACKUP};
            // Default state for server instances is BACKUP, in this state:
            //                  - listen for Master heartbeats, and update propagations
            //                  - do not accept client connections
            //                  - elevate to MASTER state if three consecutive master heartbeats missed

            // MASTER state, in this state:
            //                  - send periodic Master heartbeats (broadcast)
            //                  - listen for Master heartbeats
            //                  - accept client connections
            //                  - demote to BACKUP state if master heartbeat from another server instance is detected

    public enum MessageType {SERVER_ADVERTISEMENT, HEARTBEAT, REQUEST_RECORD, SEND_RECORD, 
        DATA_UPDATE_FROM_CLIENT, UPDATE_PROPAGATION_FROM_MASTER, CLIENT_DISCONNECT};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Application_Message_PDU 
    {
            MessageType eMessageType;   // Size 4 Offset 0
            int iSender_Random_ID;      // Size 4 Offset 4      Needed so that UPDATE_PROPAGATION_FROM_MASTER messages from self can be detected and ignored
            int iCust_ID;               // Size 4 Offset 8
            Boolean bCustNameValid;     // Size 1 Offset 12     Single byte but will need to pad to 4-byte boundary (i.e. add 3 bytes padding)
            byte[] cCustName;           // Size MAX_CUST_NAME_LEN + 1 (21) Offset 16	
    };

    private static int Application_Message_Size = 36;

    final class ServerAdvertisement_Message_PDU
    {
        MessageType eMessageType;   // Size 4 Offset 0
        int iSender_Random_ID;      // Size 4 Offset 4			
        int iAddressFamily;         // Size 4 Offset 8
        int iPort;                  // Size 4 Offset 10-11
        byte iIP_Address_Byte_1;    // Size 1 Offset 12
        byte iIP_Address_Byte_2;    // Size 1 Offset 13
        byte iIP_Address_Byte_3;    // Size 1 Offset 14
        byte iIP_Address_Byte_4;    // Size 1 Offset 15
    };
    
    private static int ServerAdvertisement_Message_Size = 24;

    DatagramSocket m_ServerAdvertisementsRecvSocket;
    Socket m_ConnectSocket;
    InputStream m_InStream;
    OutputStream m_OutStream;
    InetAddress m_Local_Address;
    InetAddress m_Server_Address;
    int m_iServerPort;
    
    Boolean m_bConnected;

    Timer m_Timer;
    
    public Replication_Client()
    {
        initComponents();
        GetLocalIPAddress();
        m_bConnected = false;
        InitialiseControls();
        Create_ReceiveServerAdvertisementsSocket();
    }

    private void InitialiseControls()    
    {    
        jTextField_Server_IPAddress.setEnabled(false);
        jTextField_Server_IPAddress.setText("-");
        jTextField_Server_Port.setEnabled(false);
        jTextField_Server_Port.setText("-");
        jButton_Connect.setEnabled(false); // Do not enable the connect button until a service advertisement message has been received 
        jTextField_CustomerID.setEnabled(false);
        jButton_RequestData.setEnabled(false); 
        jTextField_CustomerName.setEnabled(false);
        jButton_SendUpdate.setEnabled(false); 
        jButton_ClearLog.setEnabled(true); 
        jButton_Done.setEnabled(true); 
        jTextArea_EventLog.setEnabled(false);
        WriteEventLogEntry("Waiting for Server Advertisement message");    
    }

    void GetLocalIPAddress()
    {   
       String sHostAddress = "";
        try
        {
            m_Local_Address = InetAddress.getLocalHost();
            sHostAddress = m_Local_Address.toString(); // Expected format: "Hostname/192.168.100.5"
            String sAddress = String.format("Local Address %s", sHostAddress);
            WriteEventLogEntry(sAddress);
            String sStr = String.format("Server Advertisements received on port %d", REPLICATION_SERVER_ADVERTISEMENT_PORT);
            WriteEventLogEntry(sStr);
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"DB_Arch_AppServer_for_DB1",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
    }
        
    void GetServerAddress_FromServerAdvertisementMessage(ServerAdvertisement_Message_PDU ReceivedMessage)
    {	
        // Extract the IP address and port number from the received Server Advertisement Message
        byte[] bAddress = new byte[4];
        bAddress[0] = ReceivedMessage.iIP_Address_Byte_1;
        bAddress[1] = ReceivedMessage.iIP_Address_Byte_2;
        bAddress[2] = ReceivedMessage.iIP_Address_Byte_3;
        bAddress[3] = ReceivedMessage.iIP_Address_Byte_4;
        try 
        {
            m_Server_Address = InetAddress.getByAddress(bAddress);
        } 
        catch (UnknownHostException Ex)
        {
            WriteEventLogEntry(Ex.toString());
        }
        jTextField_Server_IPAddress.setText(DottedIPAddressString_from_4ByteArray(bAddress));
        m_iServerPort = ReceivedMessage.iPort;
        jTextField_Server_Port.setText(Integer.toString(m_iServerPort));
        String sStr = String.format("Received Server Advertisement IP: %s  Port:%d", 
                    DottedIPAddressString_from_4ByteArray(bAddress), m_iServerPort);
        WriteEventLogEntry(sStr);
    }

    void Create_ReceiveServerAdvertisementsSocket()
    {
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_ServerAdvertisementsRecvSocket = new DatagramSocket(REPLICATION_SERVER_ADVERTISEMENT_PORT); // create socket and bind to the port for receiveing
            m_ServerAdvertisementsRecvSocket.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Client",0);
           CloseAndExit();
        }

        StartTimer();   // As all activity is non-blocking
    }

    void CreateAndConnectSocket()
    {
        try
        {
            m_ConnectSocket = new Socket(m_Server_Address, m_iServerPort); // Create Socket and connect to Server
            m_InStream = m_ConnectSocket.getInputStream();
            m_OutStream = m_ConnectSocket.getOutputStream();
            m_ConnectSocket.setSoTimeout(100); // Timeout = 100ms  (i.e non-blocking IO behaviour)
        }  
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),
               "Is server address/port correct? Is server running?",
               JOptionPane.PLAIN_MESSAGE);
           return;
        }
        m_bConnected = true;
        jTextField_CustomerID.setEnabled(true);
        jButton_RequestData.setEnabled(true);
        jTextField_Server_IPAddress.setEnabled(false);
        jButton_Connect.setEnabled(false);
        jButton_RequestData.setEnabled(true);
        WriteEventLogEntry("Connected to Server");
    }

    void StartTimer()
    {
        m_Timer = new Timer();
        m_Timer.scheduleAtFixedRate(
                new OnTimedEvent_Timer(), 100, 100); // Short interval to allow UI responsiveness)
    }

    void StopTimer()
    {
        if (null != m_Timer)
        {
            m_Timer.cancel();
        }
    }
    
    private class OnTimedEvent_Timer extends TimerTask
    {
        public void run()
        {
            DoReceiveServerAdvertisement();
            DoReceiveApplicationMessage();
        }
    }

    void DoReceiveServerAdvertisement()
    {   // Check for Server Advertisement messages
        byte[] bArray = new byte[ServerAdvertisement_Message_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bArray, ServerAdvertisement_Message_Size);
        try
        {
            m_ServerAdvertisementsRecvSocket.receive(RecvPacket);
        }
        catch (SocketTimeoutException Ex)
        {
            return;
        }
        catch (Exception Ex)
        {
            return;
        }
        if(ServerAdvertisement_Message_Size <= RecvPacket.getLength())
        {
            ServerAdvertisement_Message_PDU Message = DeSerialize_ServerAdvertisementMessage(bArray);
            switch(Message.eMessageType)
            {
                case SERVER_ADVERTISEMENT:
                    if (false == m_bConnected)
                    {
                        GetServerAddress_FromServerAdvertisementMessage(Message);
                        jButton_Connect.setEnabled(true);
                        WriteEventLogEntry("Received SERVER_ADVERTISEMENT message");
                    }
                    break;
            }
        }
    }
    
    void DoReceiveApplicationMessage()
    {   // Check for application messages
        byte[] bArray = new byte[Application_Message_Size];
        int iReceiveByteCount = 0;
        if (true == m_bConnected)
        {
            try
            {
                iReceiveByteCount = m_InStream.read(bArray);
            }
            catch (SocketTimeoutException Ex)
            {
                return;
            }
            catch (Exception Ex)
            {
                return;
            }
            
            if(Application_Message_Size <= iReceiveByteCount)
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
                switch (Message.eMessageType)
                {
                    case SEND_RECORD:
                        String sStr = new String(Message.cCustName);
                        String sCust = TruncateString(sStr, sStr.length());
                        sStr = String.format("Received SEND_RECORD message: %s", sCust);
                        WriteEventLogEntry(sStr);
                        jTextField_CustomerName.setText(sCust);
                        jTextField_CustomerName.setEnabled(true); // Allow user to update the value
                        jButton_SendUpdate.setEnabled(true);
                        break;
                }
            }
        }
    }

    ServerAdvertisement_Message_PDU DeSerialize_ServerAdvertisementMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
            ServerAdvertisement_Message_PDU result_Message = new ServerAdvertisement_Message_PDU();
            result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
            byte[] bInt = new byte[4]; 
            System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
            result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);            
            result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
            // Deserialisation of port number
            // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            result_Message.iPort = (256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/;
            result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
            result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
            result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
            result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
            return result_Message;
        }

    Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
            byte[] bInt = new byte[4]; 
            System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
            result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);
            System.arraycopy(bRecdByteArray/*src*/, 8, bInt/*dest*/, 0, 4/*count*/); 
            result_Message.iCust_ID = ByteArraySize4_ToInt(bInt);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[12])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
            System.arraycopy(bRecdByteArray/*src*/, 13, result_Message.cCustName/*dest*/, 0, MAX_CUST_NAME_LEN/*count*/); 
            return result_Message;
        }

    int GetCustIDFromControl()
    {
        int iCustID = 0;
        if ("" == jTextField_CustomerID.getText())
        {
            JOptionPane.showMessageDialog(null,
                "Customer ID must be between 101 and 103", "Invalid Customer ID",
                JOptionPane.PLAIN_MESSAGE);
            return -1;
        }
        try
        {
            iCustID = Integer.parseInt(jTextField_CustomerID.getText());
            if (iCustID < 101 || iCustID > 103)
            {
            JOptionPane.showMessageDialog(null,
                "Customer ID must be between 101 and 103", "Invalid Customer ID",
                JOptionPane.PLAIN_MESSAGE);
                return -1;
            }
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null, Ex.toString(), "Invalid Customer ID",
                JOptionPane.PLAIN_MESSAGE);
            return -1;
        }
        return iCustID;
    }    
         
        private void RequestData()
        {
            int iCustID = GetCustIDFromControl();
            if (-1 == iCustID)
            {
                return;
            }
            jTextField_CustomerName.setText("");   // Initially clear the data control
            Application_Message_PDU Message = new Application_Message_PDU();
            Message.eMessageType = MessageType.REQUEST_RECORD;
            Message.iCust_ID = iCustID;
            Message.bCustNameValid = false;
            if (true == m_bConnected)
            {
                SendApplicationMessage(Message);  // Send REQUEST_RECORD message to AppServer
                WriteEventLogEntry("Sent REQUEST_RECORD message");
            }
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            bByteArray[0] = (byte)Message.eMessageType.ordinal();
            byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
            System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
            bInt = IntToByteArray(Message.iCust_ID);
            System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 8, 4/*count*/);
            bByteArray[12] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[12] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                System.arraycopy(Message.cCustName/*src*/, 0, bByteArray/*dest*/, 13, Message.cCustName.length/*count*/);
            }
            return bByteArray;
        }

        private void SendUpdate()
        {
            int iCustID = GetCustIDFromControl();
            if (-1 == iCustID)
            {
                return;
            }
            Application_Message_PDU Message = new Application_Message_PDU();
            Message.eMessageType = MessageType.DATA_UPDATE_FROM_CLIENT;
            Message.iCust_ID = iCustID;
            Message.cCustName = jTextField_CustomerName.getText().getBytes();
            Message.bCustNameValid = true;
            if (true == m_bConnected)
            {
                SendApplicationMessage(Message);  // Send REQUEST_RECORD message to AppServer
            }
            WriteEventLogEntry("Sent DATA_UPDATE_FROM_CLIENT message");
            jTextField_CustomerName.setEnabled(false); // Require refresh from server (client must request) before further updates can be sent
            jButton_SendUpdate.setEnabled(false); // Data must be edited before the button can be pressed
        }

    void SendApplicationMessage(Application_Message_PDU Message)
    {
        byte[] bBuffer = Serialize_Application_Message_PDU(Message);
        int iBytesSent;
        try
        {
            m_OutStream.write(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
            return;
        }
        catch (Exception Ex)
        {
            return;
        }
    }

    void CloseSockets()
    {
        try
        {
            m_ConnectSocket.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_ServerAdvertisementsRecvSocket.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
       if(null != m_Timer)
        {
            m_Timer.cancel();
        }
        CloseSockets();
        System.exit(0);
    }

    String DottedIPAddressString_from_4ByteArray(byte[] bAddress)
    {
        String sStr = "";
        int iComponent = 0;
        if(4 != bAddress.length)
        {
            return sStr;    // If byte array is not 4 bytes return empty string
        }
        iComponent = bAddress[0] & 0xFF;
        sStr = Integer.toString(iComponent) + ".";
        iComponent = bAddress[1] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[2] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[3] & 0xFF;            
        sStr += Integer.toString(iComponent);
        return sStr;
    }
    
    private byte[] DottedIPAddessString_To_ByteArraySize4(String sDottedIPAddress)
    {
        byte[] bAddress = new byte[4]; // Each byte initialised as value 0
        int iLength = sDottedIPAddress.length();
        if(7 > iLength || 15 < iLength) // A well formed dotted IP address string must be in this length range
        {
            return bAddress; // All 0's signals mal-formed dotted IPaddress string
        }
        // sDottedIPAddress is a '.' separated, list of numeric values
        // There is no '.' after the last item in the list
        int iNextDotPos  = 0;
        String sCurrentValue = "";
        int iArrayIndex = 0;
        try
        {
            while ((iNextDotPos = sDottedIPAddress.indexOf('.')) > 0 && iNextDotPos < iLength) // Ignore chars beyond the valid string content
            {
                sCurrentValue = sDottedIPAddress.substring(0/*start*/, iNextDotPos/*length*/);
                bAddress[iArrayIndex++] = (byte)Integer.parseInt(sCurrentValue);
                sDottedIPAddress = sDottedIPAddress.substring(iNextDotPos + 1/*start*/);// Remove 'dealt with' name from head of list
                iLength -= (iNextDotPos + 1);                                 // Keep track of remaining list-string length
            }
            // Deal with residual substring (there is no comma after the last item in the list)
            bAddress[iArrayIndex++] = (byte)Integer.parseInt(sDottedIPAddress);
        }
        catch (Exception Ex)
        {   // Catch various string-format related exceptions (such as having more than three '.'s which
            // would cause the byte array index to go out of bounds)
            bAddress[0] = 0; // signal mal-formed dotted IPaddress string 
        }
        return bAddress;
    }
    
    private static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    private int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }

    private String TruncateString(String sStr, int iMaxLength)
    {
        try
        {
            // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
            int iNullTerminatePosition = sStr.indexOf(0);
            if(-1 != iNullTerminatePosition)
            {
                sStr = sStr.substring(0, iNullTerminatePosition);
            }
            if(sStr.length()> iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.substring(0, iMaxLength);
            }
            return sStr;
        }
        catch (Exception Ex)
        {
            return "";
        }
    }
   
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        jTextArea_EventLog.append(sStr);       
    }

    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new Replication_Client().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_Connect == e.getSource())
        {
            CreateAndConnectSocket();
        }
        if(jButton_RequestData == e.getSource())
        {
            RequestData();
        }
        if(jButton_SendUpdate == e.getSource())
        {
            SendUpdate();
        }
        if(jButton_ClearLog == e.getSource())
        {
            jTextArea_EventLog.setText("");
        }
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
    }
    
    private JPanel jPanel_ConnectConfiguration;
    private JLabel jLabel_Server_IPAddress;
    private JTextField jTextField_Server_IPAddress;
    private JLabel jLabel_Server_Port;
    private JTextField jTextField_Server_Port;
    private JButton jButton_Connect;
    private JPanel jPanel_Controls;
    private JLabel jLabel_CustomerID1;
    private JLabel jLabel_CustomerID2;
    private JTextField jTextField_CustomerID;
    private JButton jButton_RequestData;
    private JLabel jLabel_CustomerName;
    private JTextField jTextField_CustomerName;
    private JButton jButton_SendUpdate;
    private JButton jButton_ClearLog;
    private JButton jButton_Done;
    private JPanel jPanel_EventLog;
    private JTextArea jTextArea_EventLog;
    private JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_Server_IPAddress = new JLabel();
        jLabel_Server_IPAddress.setText("IP Address");
        jTextField_Server_IPAddress = new JTextField();
        jTextField_Server_IPAddress.setMaximumSize(new Dimension(110, 20));
        jTextField_Server_IPAddress.setHorizontalAlignment(JTextField.CENTER);
        jLabel_Server_Port = new JLabel();
        jLabel_Server_Port.setText("Port");
        jTextField_Server_Port = new JTextField();
        jTextField_Server_Port.setMaximumSize(new Dimension(50, 20));
        jTextField_Server_Port.setHorizontalAlignment(JTextField.CENTER);
        jButton_Connect = new javax.swing.JButton();
        jButton_Connect.setText("Connect");
        jButton_Connect.setMaximumSize(new Dimension(50, 4));
        jButton_Connect.addActionListener(this);
        jPanel_ConnectConfiguration = new JPanel();
        jPanel_ConnectConfiguration.setPreferredSize(new Dimension(550, 100));
        jPanel_ConnectConfiguration.setBorder(BorderFactory.createTitledBorder("Application Server   Connect configuration"));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_ConnectConfiguration);
        jPanel_ConnectConfiguration.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel1Layout.createSequentialGroup()
                    .addContainerGap(30, 30)
                    .add(jLabel_Server_IPAddress)
                    .addContainerGap(20, 20)
                    .add(jTextField_Server_IPAddress)
                    .addContainerGap(70, 70)
                    .add(jLabel_Server_Port)
                    .addContainerGap(20, 20)
                    .add(jTextField_Server_Port)
                    .addContainerGap(70, 70)
                    .add(jButton_Connect)
                    .addContainerGap(30, 30)));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jLabel_Server_IPAddress)
                .add(jTextField_Server_IPAddress)
                .add(jLabel_Server_Port)
                .add(jTextField_Server_Port)
                .add(jButton_Connect));

        jLabel_CustomerID1 = new JLabel();
        jLabel_CustomerID1.setText("Customer ID");
        jLabel_CustomerID2 = new JLabel();
        jLabel_CustomerID2.setText("(use values 101, 102 or 103)");
        jTextField_CustomerID = new JTextField();
        jTextField_CustomerID.setMaximumSize(new Dimension(50, 20));
        jTextField_CustomerID.setHorizontalAlignment(JTextField.CENTER);
        jButton_RequestData = new javax.swing.JButton();
        jButton_RequestData.setText("Request data");
        jButton_RequestData.setMaximumSize(new Dimension(30, 4));
        jButton_RequestData.addActionListener(this);
        jLabel_CustomerName = new JLabel();
        jLabel_CustomerName.setText("Customer name");
        jTextField_CustomerName = new JTextField();
        jTextField_CustomerName.setMaximumSize(new Dimension(100, 20));
        jTextField_CustomerName.setHorizontalAlignment(JTextField.CENTER);
        jButton_SendUpdate = new javax.swing.JButton();
        jButton_SendUpdate.setText("Send updated data");
        jButton_SendUpdate.setMaximumSize(new Dimension(30, 4));
        jButton_SendUpdate.addActionListener(this);
        jButton_ClearLog = new javax.swing.JButton();
        jButton_ClearLog.setText("Clear Log entries");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new javax.swing.JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls = new javax.swing.JPanel();
        jPanel_Controls.setPreferredSize(new Dimension(200, 300));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder("Data and Controls"));

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerID1)
                    .addContainerGap(120, 120))
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerID2)
                    .addContainerGap(35, 35))
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jTextField_CustomerID)
                    .addContainerGap(20, 20)
                    .add(jButton_RequestData))
                .add(jLabel_CustomerName)
                .add(jTextField_CustomerName)
                .add(jButton_SendUpdate)
                .add(jButton_ClearLog)
                .add(jButton_Done));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerID1)
                    .add(jLabel_CustomerID2)
                    .add(jPanel2Layout.createParallelGroup()
                        .add(jTextField_CustomerID)
                        .add(jButton_RequestData))
                    .addContainerGap(10, 10)
                    .add(jLabel_CustomerName)
                    .add(jTextField_CustomerName)
                    .addContainerGap(10, 10)
                    .add(jButton_SendUpdate)
                    .addContainerGap(40, 40)
                    .add(jButton_ClearLog)
                    .addContainerGap(10, 10)
                    .add(jButton_Done)));
        jPanel2Layout.linkSize(new java.awt.Component[] {jButton_SendUpdate, jButton_ClearLog, jButton_Done});

        jTextArea_EventLog = new javax.swing.JTextArea();
        jScrollPane_EventLog = new javax.swing.JScrollPane(jTextArea_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog = new javax.swing.JPanel();
        jPanel_EventLog.setPreferredSize(new Dimension(300, 350));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event Log (Diagnostic)"));
        
        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));        
         
        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(550, 350));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)            
                    .add(jPanel_ConnectConfiguration)
                    .add(layout.createSequentialGroup()
                        .add(jPanel_Controls)
                        .add(jPanel_EventLog))));
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)   
                .add(layout.createSequentialGroup()
                    .add(jPanel_ConnectConfiguration)
                    .add(layout.createParallelGroup()  
                    .add(jPanel_Controls)
                    .add(jPanel_EventLog))));

        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        setTitle("Replication CLIENT");
        pack();
    }                                             
}