/*	File			Replication_Server_DB_Transfer.cs
	Purpose			Replication demonstration with automatic full database transfer to newly started backup instances
	Author			Richard John Anthony
	Date			March 2015
*/

package Replication_Server_DB_Transfer;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.*;
import java.util.Random;
    
public class Replication_Server_DB_Transfer extends javax.swing.JFrame implements ActionListener
{
    private static int REPLICATION_EXTERNAL_PORT = 8006;  // Client to Server
    private static int REPLICATION_UPDATE_PORT = 8007;	// Server to Server - Update propagation
    private static int REPLICATION_HEARTBEAT_PORT = 8008; // Server to Server - Master heartbeat broadcasts
    private static int REPLICATION_SERVER_ADVERTISEMENT_PORT = 8009; // Server to client
    private static int DATABASE_SIZE = 3;
    private static int MAX_CUST_NAME_LEN = 20;

    public enum ServerState {MASTER, BACKUP};
            // Default state for server instances is BACKUP, in this state:
            //                  - listen for Master heartbeats, and update propagations
            //                  - do not accept client connections
            //                  - elevate to MASTER state if three consecutive master heartbeats missed

            // MASTER state, in this state:
            //                  - send periodic Master heartbeats (broadcast)
            //                  - listen for Master heartbeats
            //                  - accept client connections
            //                  - demote to BACKUP state if master heartbeat from another server instance is detected

    public enum MessageType {SERVER_ADVERTISEMENT, HEARTBEAT, REQUEST_RECORD, SEND_RECORD,
        DATA_UPDATE_FROM_CLIENT, UPDATE_PROPAGATION_FROM_MASTER, CLIENT_DISCONNECT, REQUEST_FULL_DATABASE_TRANSFER};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Application_Message_PDU 
    {
            MessageType eMessageType;   // Size 4 Offset 0
            int iSender_Random_ID;      // Size 4 Offset 4      Needed so that UPDATE_PROPAGATION_FROM_MASTER messages from self can be detected and ignored
            int iCust_ID;               // Size 4 Offset 8
            Boolean bCustNameValid;     // Size 1 Offset 12     Single byte but will need to pad to 4-byte boundary (i.e. add 3 bytes padding)
            byte[] cCustName;           // Size MAX_CUST_NAME_LEN + 1 (21) Offset 16	
    };

    private static int Application_Message_Size = 36;

    final class Heartbeat_Message_PDU
    {
        MessageType eMessageType;           // Size 4 Offset 0
        int iSender_Random_ID;              // Size 4 Offset 4  Needed so that messages from self can be detected and ignored
        int iSender_Random_ID_Incremented;  // Size 4 Offset 8  Used to differentiate genuine heartbeat messages from any other messages that just happen to arrive on this port
    };

    private static int Heartbeat_Message_Size = 12;

    final class ServerAdvertisement_Message_PDU
    {
        MessageType eMessageType;   // Size 4 Offset 0
        int iSender_Random_ID;      // Size 4 Offset 4			
        int iAddressFamily;         // Size 4 Offset 8
        int iPort;                  // Size 4 Offset 10-11
        byte iIP_Address_Byte_1;    // Size 1 Offset 12
        byte iIP_Address_Byte_2;    // Size 1 Offset 13
        byte iIP_Address_Byte_3;    // Size 1 Offset 14
        byte iIP_Address_Byte_4;    // Size 1 Offset 15
    };
    
    private static int ServerAdvertisement_Message_Size = 24;

    ServerSocket m_ListenSock;  // Listen for client connections
    Socket m_ConnectSock; // Connection to Client
    InputStream m_InStream;
    OutputStream m_OutStream;
    MulticastSocket m_HeartbeatBroadcastSock; // Send heartbeat broadcasts
    DatagramSocket m_HeartbeatReceiveSock; // Recieve heartbeat broadcasts
    MulticastSocket m_ServerAdvertisementBroadcastSock;
    MulticastSocket m_BroadcastPropagatedUpdateSock; // Send heartbeat broadcasts
    DatagramSocket m_ReceivePropagatedUpdateSock; // Recieve propagated database update broadcasts

    InetAddress m_Local_Address;    // For all receives
    InetAddress m_Broadcast_Address; // For broadcasting Heartbeats, Server Advertisements and Update propagations

    ServerState m_ServerState;
    int m_iHeartbeatMissedCount;
    int m_iServer_Random_ID;
    Boolean m_bConnected;
    Boolean m_bFreshlyStarted;	// Required to determine whether to request a DB transfer
    
    Timer m_Timer_Application;
    Timer m_Timer_Send_Heartbeat;
    Timer m_Timer_Receive_Heartbeat;
    Timer m_Timer_Send_ServerAdvertisement;
    Boolean m_bHeartbeatTimerRunning;
    Boolean m_bServerAdvertisementTimerRunning;

    // Data Row and Table structures (hard coded data serves as mock of real database)
    final class DataRow_struct 
    {
        int iCust_ID;
        byte[] cCustName;	
    };

    // DataTable_struct
    DataRow_struct[] m_DataRowArray; // Array of data row structures
                 
    public Replication_Server_DB_Transfer()
    {
        initComponents();

        m_DataRowArray = new DataRow_struct[DATABASE_SIZE];
        GetLocalIPAddress();

        Random rnd = new Random();
        m_iServer_Random_ID = rnd.nextInt(10000);
        jTextField_ServerID.setText(Integer.toString(m_iServer_Random_ID));

        m_bHeartbeatTimerRunning = false;
        m_bServerAdvertisementTimerRunning = false;	
        m_bConnected = false;
        m_bFreshlyStarted = true;

        InitialiseDatabase();
        DisplayDataTable();
        EnableServer();
        Initialise_Send_Heartbeat_Socket();
        Initialise_Send_ServerAdvertisement_Socket();
        Initialise_Broadcast_Update_propagation_Socket();
        Enable_ReceiveHeartbeatMessages();
        Start_Timer_Receive_Heartbeat();	// Initiate listening for heartbeat messages	
        Enable_ReceivePropagatedUpdateMessages();

        Transition_To_BACKUP_State(true);
        InitialiseControls();
    }

    private void InitialiseControls()    
    {    
        jTextArea_CustomerID.setEnabled(false);;
        jTextArea_CustomerName.setEnabled(false);
        jTextField_ServerState.setEnabled(false);
        jTextField_ServerID.setEnabled(false);
        jTextField_HeartbeatMissedCount.setEnabled(false);
        jButton_ClearLog.setEnabled(true); 
        jButton_Done.setEnabled(true); 
        jTextArea_EventLog.setEnabled(false);
    }

    void GetLocalIPAddress()
    {   
        String sHostAddress = "";
        try
        {
            m_Local_Address = InetAddress.getLocalHost();
            sHostAddress = m_Local_Address.toString(); // Expected format: "Hostname/192.168.100.5"
            String sAddressDisplay = "Local address " + sHostAddress;
            WriteEventLogEntry(sAddressDisplay);            
            String sStr = String.format("Send / Recv application messages port %d",REPLICATION_EXTERNAL_PORT);
            WriteEventLogEntry(sStr);

            m_Broadcast_Address = GetLocalBroadcastAddress_From_LocalHostAddress(m_Local_Address);
            byte[] bAddress = m_Broadcast_Address.getAddress();
            WriteEventLogEntry("B'cast address for broadcasting Heartbeats, Server");
            sStr = String.format("     Advertisements and Update Propagation %s", DottedIPAddressString_from_4ByteArray(bAddress));
            WriteEventLogEntry(sStr);

            sStr = String.format("B'cast / Recv update propagation on port %d",REPLICATION_UPDATE_PORT);
            WriteEventLogEntry(sStr);
            sStr = String.format("B'cast / Recv heartbeats on port %d",REPLICATION_HEARTBEAT_PORT);
            WriteEventLogEntry(sStr);
            sStr = String.format("B'cast / Recv server advertisements on port %d",REPLICATION_SERVER_ADVERTISEMENT_PORT);
            WriteEventLogEntry(sStr);
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
    }

    void InitialiseDatabase()	// Populate DataTable (hardcoded data)
    {
        m_DataRowArray = new DataRow_struct[DATABASE_SIZE];
        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {
            m_DataRowArray[iIndex] = new DataRow_struct();
            m_DataRowArray[iIndex].cCustName = new byte[MAX_CUST_NAME_LEN+1];
        }
        m_DataRowArray[0].iCust_ID = 101;
        String sCustName = "Fred";
        m_DataRowArray[0].cCustName = sCustName.getBytes();
        m_DataRowArray[1].iCust_ID = 102;
        sCustName = "Jane";
        m_DataRowArray[1].cCustName = sCustName.getBytes();
        m_DataRowArray[2].iCust_ID = 103;
        sCustName = "Rita";
        m_DataRowArray[2].cCustName = sCustName.getBytes();
    }

    void DisplayDataTable()
    {	
        jTextArea_CustomerID.setText("");
        jTextArea_CustomerName.setText("");
        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex ++)
        {
            jTextArea_CustomerID.append(Integer.toString(m_DataRowArray[iIndex].iCust_ID) + "\n");
            jTextArea_CustomerName.append(new String(m_DataRowArray[iIndex].cCustName) + "\n");
        }
    }

    void UpdateDatabase(int iCustID_AsIndex, String sCustName)	// Upate DataTable
    {
        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {
            if(iCustID_AsIndex == m_DataRowArray[iIndex].iCust_ID)
            {
                String sStr = TruncateString(sCustName, MAX_CUST_NAME_LEN);
                m_DataRowArray[iIndex].cCustName = sStr.getBytes();
                DisplayDataTable();
                return;
            }
        }		
    }
         
    void Accept()
    {
        try
        {
            m_ConnectSock = m_ListenSock.accept();
            m_ConnectSock.setSoTimeout(100); // 100ms timeout (non-blocking)            }
            m_InStream = m_ConnectSock.getInputStream();
            m_OutStream = m_ConnectSock.getOutputStream();
        }
        catch (SocketTimeoutException Ex)
        {   
            return;
        }
        catch (Exception Ex)
        {
            WriteEventLogEntry(Ex.toString());
            return;
        }

        m_bConnected = true;
        String sStr = String.format("Client %s Connected", m_ConnectSock.getRemoteSocketAddress().toString());
        WriteEventLogEntry(sStr);
    }

    void EnableServer()  // Invoked on startup
    {
        try
        {   // Create, Bind and Listen on Listen Socket and set it NonBlocking
            m_ListenSock = new ServerSocket(REPLICATION_EXTERNAL_PORT, 5/*backlog*/, m_Local_Address/*bindAddr*/);
            m_ListenSock.setSoTimeout(100); // Timeout = 100mS i.e. non-blocking
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
        Start_Timer_Application();   // As all activity is non-blocking
    }

    void Enable_ReceiveHeartbeatMessages()  // Invoked on startup
    {
        try
        {   // Create and bind Socket, set it NonBlocking
            m_HeartbeatReceiveSock = new DatagramSocket(REPLICATION_HEARTBEAT_PORT);    // create socket and bind to the port for receiveing
            m_HeartbeatReceiveSock.setSoTimeout(100);   // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
    }

    void Enable_ReceivePropagatedUpdateMessages()  // Invoked on startup
    {
        try
        {   // Create and bind Socket, set it NonBlocking
            m_ReceivePropagatedUpdateSock = new DatagramSocket(REPLICATION_UPDATE_PORT);    // create socket and bind to the port for receiveing
            m_ReceivePropagatedUpdateSock.setSoTimeout(100);   // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
    }

    void Transition_To_MASTER_State()   // replace all other transitions !
    {
        Start_Timer_Send_ServerAdvertisement(); // Start timer to govern sending of Server Advertisement messages
        Start_Timer_Send_Heartbeat();           // Start timer to govern sending of Heartbeat messages
        m_ServerState = ServerState.MASTER;
        Update_ServerState_Display();
        m_iHeartbeatMissedCount = 0;
        jTextField_HeartbeatMissedCount.setText(Integer.toString(m_iHeartbeatMissedCount));
        WriteEventLogEntry("Promoted to MASTER State (no other MASTER detected)");
    }

    void Transition_To_BACKUP_State(Boolean bSilent /*dont display 'demoted' message when called during initialisation*/)
    {
        if(true == m_bHeartbeatTimerRunning)
        {
            m_Timer_Send_Heartbeat.cancel();    // Stop sending periodic heartbeats
            m_bHeartbeatTimerRunning = false;
        }

        if(true == m_bServerAdvertisementTimerRunning)
        {
            m_Timer_Send_ServerAdvertisement.cancel();  // Stop sending Server Advertisements
            m_bServerAdvertisementTimerRunning = false;
        }

        m_ServerState = ServerState.BACKUP;	// Default state is backup - in this state listen for Master heartbeats 
        Update_ServerState_Display();
        m_iHeartbeatMissedCount = 0;
        jTextField_HeartbeatMissedCount.setText(Integer.toString(m_iHeartbeatMissedCount));

        if(false == bSilent)
        {
            WriteEventLogEntry("Demoted to BACKUP status (other MASTER detected)");
        }
    }
    
    private void Start_Timer_Application()
    {
        m_Timer_Application = new Timer();
        m_Timer_Application.scheduleAtFixedRate(
                new OnTimedEvent_Timer_Application(), 100, 100);
    }

    private class OnTimedEvent_Timer_Application extends TimerTask
    {
        public void run()
        {
            DoReceive_Heartbeat(); // Check for receipt of Heartbeat messages
            DoReceive();
            if(false == m_bConnected)
            {
                Accept();
            }
        }
    }
    
    private void Start_Timer_Send_Heartbeat()
    {
        m_Timer_Send_Heartbeat = new Timer();
        m_Timer_Send_Heartbeat.scheduleAtFixedRate(
                new OnTimedEvent_Timer_Send_Heartbeat(), 3000, 3000);
    }

    private class OnTimedEvent_Timer_Send_Heartbeat extends TimerTask
    {
        public void run()
        {
            if(ServerState.BACKUP == m_ServerState)
            {
                // No action, BACKUP server instances do not send Heartbeats
            }
            if(ServerState.MASTER == m_ServerState)
            {
                WriteEventLogEntry("Send heartbeat");
                Send_Heartbeat();   // Broadcast heartbeat message		
            }
        }
    }
    
    private void Start_Timer_Receive_Heartbeat()
    {
        m_Timer_Receive_Heartbeat = new Timer();
        m_Timer_Receive_Heartbeat.scheduleAtFixedRate(
                new OnTimedEvent_Timer_Receive_Heartbeat(), 3300, 3300);// 3.3 second interval between
                                // checking for heartbeat messages (i.e longer than sending interval)
    }

    private class OnTimedEvent_Timer_Receive_Heartbeat extends TimerTask
    {
        public void run()
        {
            if(ServerState.BACKUP == m_ServerState)
            {
                m_iHeartbeatMissedCount ++;
                jTextField_HeartbeatMissedCount.setText(Integer.toString(m_iHeartbeatMissedCount));
                String sStr = String.format("Heartbeat missed! count = %d", m_iHeartbeatMissedCount);
                WriteEventLogEntry(sStr);
                if(3 == m_iHeartbeatMissedCount)
                {
                    Transition_To_MASTER_State();
                }
            }
            if(ServerState.MASTER == m_ServerState)
            {
                // No action as Heartbeat was not expected, so its absence requires no action
            }
        }
    }

    private void Start_Timer_Send_ServerAdvertisement()
    {
        m_Timer_Send_ServerAdvertisement = new Timer();
        m_Timer_Send_ServerAdvertisement.scheduleAtFixedRate(
                new OnTimedEvent_Timer_Send_ServerAdvertisement(), 5000, 5000); // 5 second interval between sending ServerAdvertisement messages
    }

    private class OnTimedEvent_Timer_Send_ServerAdvertisement extends TimerTask
    {
        public void run()
        {
            if(ServerState.BACKUP == m_ServerState)
            {
                // No action, BACKUP server instances do not send Server Advertisements
            }
            if(ServerState.MASTER == m_ServerState)
            {
                WriteEventLogEntry("Send Server Advertisement");
                Send_ServerAdvertisement(); // Broadcast Server Advertisement message					
            }
        }
    }
    
    void Initialise_Send_ServerAdvertisement_Socket()
    {
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_ServerAdvertisementBroadcastSock = new MulticastSocket(); // Create socket and bind to any available port
            m_ServerAdvertisementBroadcastSock.setBroadcast(true);     // Enable broadcast sends
            m_ServerAdvertisementBroadcastSock.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
    }

    void Initialise_Send_Heartbeat_Socket()
    {
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_HeartbeatBroadcastSock = new MulticastSocket(); // Create socket and bind to any available port
            m_HeartbeatBroadcastSock.setBroadcast(true);     // Enable broadcast sends
            m_HeartbeatBroadcastSock.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
    }

    void Initialise_Broadcast_Update_propagation_Socket()
    {
        try
        {   // Create and bind Socket, set it NonBlocking and enable broadcast sends
            m_BroadcastPropagatedUpdateSock = new MulticastSocket(); // Create socket and bind to any available port
            m_BroadcastPropagatedUpdateSock.setBroadcast(true);     // Enable broadcast sends
            m_BroadcastPropagatedUpdateSock.setSoTimeout(100);      // Timeout = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Replication Server",0);
           CloseAndExit();
        }
    }
    
    void Send_Heartbeat()
    {
        Heartbeat_Message_PDU Heartbeat_Message = new Heartbeat_Message_PDU();
        Heartbeat_Message.eMessageType = MessageType.HEARTBEAT;
        Heartbeat_Message.iSender_Random_ID = m_iServer_Random_ID;
        Heartbeat_Message.iSender_Random_ID_Incremented = m_iServer_Random_ID + 1;
        DoSend_Heartbeat(Heartbeat_Message);
    }

    void Send_DatabaseTransferRequest()
    {	// Backup instance requests DB transfer from Master (upon Backup startup)
        // Re-uses the Heartbeat messages mechanism - Sends as a broadcast on the same socket as for Heartbeat messages
        // Non-Master nodes will ignore this request
        Heartbeat_Message_PDU Heartbeat_Message = new Heartbeat_Message_PDU();
        Heartbeat_Message.eMessageType = MessageType.REQUEST_FULL_DATABASE_TRANSFER;	// New message type added to support DB transfer requests
        Heartbeat_Message.iSender_Random_ID = m_iServer_Random_ID;
        Heartbeat_Message.iSender_Random_ID_Incremented = m_iServer_Random_ID + 1;
        DoSend_Heartbeat(Heartbeat_Message);
    }
    
    void DoSend_Heartbeat(Heartbeat_Message_PDU Heartbeat_Message)
    {	
        byte[] bSendBuf = Serialize_Heartbeat_Message_PDU(Heartbeat_Message);
        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                                    m_Broadcast_Address, REPLICATION_HEARTBEAT_PORT); 
            m_HeartbeatBroadcastSock.send(SendPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return;
        }
    }

    void Send_ServerAdvertisement()
    {	
        ServerAdvertisement_Message_PDU ServerAdvertisement_Message = new ServerAdvertisement_Message_PDU();
        ServerAdvertisement_Message.eMessageType = MessageType.SERVER_ADVERTISEMENT;
        ServerAdvertisement_Message.iSender_Random_ID = m_iServer_Random_ID;

        byte[] bAddress = m_Local_Address.getAddress();
        ServerAdvertisement_Message.iIP_Address_Byte_1 = bAddress[0];
        ServerAdvertisement_Message.iIP_Address_Byte_2 = bAddress[1];
        ServerAdvertisement_Message.iIP_Address_Byte_3 = bAddress[2];
        ServerAdvertisement_Message.iIP_Address_Byte_4 = bAddress[3];
        ServerAdvertisement_Message.iAddressFamily = 2; // Equivallent to C++ AF_INET     
        ServerAdvertisement_Message.iPort = REPLICATION_EXTERNAL_PORT;    

        byte[] bSendBuf = Serialize_ServerAdvertisement_Message_PDU(ServerAdvertisement_Message);
        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                        m_Broadcast_Address, REPLICATION_SERVER_ADVERTISEMENT_PORT); 
            m_ServerAdvertisementBroadcastSock.send(SendPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return;
        }
    }
    
    void SendApplicationMessage(Application_Message_PDU Message)
    {
        byte[] bBuffer = Serialize_Application_Message_PDU(Message);
        try
        {
            m_OutStream.write(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
        }
        catch (Exception ex)
        {
        }
    }

    void Send_DataRow_Message(int iCustID)
    {
        Application_Message_PDU Message = new Application_Message_PDU();
        Message.eMessageType = MessageType.SEND_RECORD;
        Message.iCust_ID = iCustID;
        String sCustName = "------No Match------";
        Message.cCustName = sCustName.getBytes();

        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {
            if(iCustID == m_DataRowArray[iIndex].iCust_ID)
            {
                Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
                System.arraycopy(m_DataRowArray[iIndex].cCustName/*src*/, 0, Message.cCustName/*dest*/, 0, m_DataRowArray[iIndex].cCustName.length/*count*/); 
            }
        }

        Message.bCustNameValid = true;
        SendApplicationMessage(Message);
        String sCust = new String(Message.cCustName);
        sCust = TruncateString(sCust, sCust.length());        
        String sStr = String.format("Data Record %d %s sent", Message.iCust_ID, sCust);
        WriteEventLogEntry(sStr);
    }

    void Send_Database_ToBackupInstance()
    {	// The Master instance transfers the DB to the Backup instance to synchronise DB copies
        // The new mechanism reuses the existing update propagation mechanism (but within a loop) to transfer the database rows one-by-one.
        // Note that the mechanism used here broadcasts the updates, which means that other Backup copies will also receive them.
        // This has the potential benefit of improving the synchronisation of the DB contents of other Backup instances but at the cost of
        // greater overheads (the Backup copies are interrupted). If the overheads are problematic, this mechanism could be modified so that
        // the database propagation is directed (point-to-point) only to the specific requesting Backup instance.

        for (int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
        {   // Loop iterates through the entire database, each time sending a single database record to the BACKUP server
            PropagateDatabaseUpdate_To_BACKUP_ServerInstance(m_DataRowArray[iIndex].iCust_ID, m_DataRowArray[iIndex].cCustName);
        }
    }
    
    void DoReceive()
    {   // Check for Database update propagation messages (a Application_Message_PDU from MASTER instance of service)
        byte[] bRecvBuf = new byte[Application_Message_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, Application_Message_Size);
        try
        {
            m_ReceivePropagatedUpdateSock.receive(RecvPacket);
            if (Application_Message_Size <= RecvPacket.getLength())
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bRecvBuf);
                switch(Message.eMessageType)
                {
                    case UPDATE_PROPAGATION_FROM_MASTER:
                        if(m_iServer_Random_ID != Message.iSender_Random_ID)
                        {   // Ignore echos from self
                            String sCust = new String(Message.cCustName);
                            String sStr = String.format("Propagated data update recd from MASTER server: %d, %s", Message.iCust_ID,sCust);
                            WriteEventLogEntry(sStr);
                            UpdateDatabase(Message.iCust_ID, sCust);
                        }
                        break;
                }
            }
        }
        catch(SocketTimeoutException Ex)
        {
        }
        catch(Exception Ex)
        {
            System.out.printf("Receive failed: %s\n", Ex.toString());
        }

	// Check for Application messages (from Client)
        byte[] bArray = new byte[Application_Message_Size];
        int iReceiveByteCount = 0;
        if (true == m_bConnected)
        {
            iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_InStream.read(bArray);
            }
            catch (SocketTimeoutException Ex)
            {
            }
            catch (Exception Ex)
            {
            }
            if(Application_Message_Size <= iReceiveByteCount)
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
                String sStr;
                switch(Message.eMessageType)
                {
                   case REQUEST_RECORD:
                        sStr = String.format("Data record %d requested", Message.iCust_ID);
                        WriteEventLogEntry(sStr);
                        Send_DataRow_Message(Message.iCust_ID);
                        break;		
                    case DATA_UPDATE_FROM_CLIENT:
                        String sCust = new String(Message.cCustName);
                        sStr = String.format("Client sent data update %d, %s",Message.iCust_ID, sCust);
                        WriteEventLogEntry(sStr);
                        UpdateDatabase(Message.iCust_ID, sCust);
                        PropagateDatabaseUpdate_To_BACKUP_ServerInstance(Message.iCust_ID, Message.cCustName);   // Forward the database update to the BACKUP server
                        break;
                    case CLIENT_DISCONNECT:
                        m_bConnected = false;
                        CloseConnectionSocket();
                        WriteEventLogEntry("Client disconnected");
                        break;
                }
            }
        }
    }

    void PropagateDatabaseUpdate_To_BACKUP_ServerInstance(int iCustID, byte[] cCustName)   // Forward the database update to the BACKUP server
    {
        Application_Message_PDU Message = new Application_Message_PDU();
        Message.eMessageType = MessageType.UPDATE_PROPAGATION_FROM_MASTER;
        Message.iSender_Random_ID = m_iServer_Random_ID;
        Message.iCust_ID = iCustID;
        Message.cCustName = new byte[MAX_CUST_NAME_LEN + 1];
        System.arraycopy(cCustName/*src*/, 0, Message.cCustName/*dest*/, 0, cCustName.length);
        Message.bCustNameValid = true;

        byte[] bSendBuf = Serialize_Application_Message_PDU(Message);
        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                                            m_Broadcast_Address, REPLICATION_UPDATE_PORT); 
            m_BroadcastPropagatedUpdateSock.send(SendPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return;
        }

        String sCust = new String(Message.cCustName);
        String sStr = String.format("Data propagation update sent: %d %s", Message.iCust_ID, sCust);
        WriteEventLogEntry(sStr);
    }

    void DoReceive_Heartbeat()
    {	// Check for heartbeat messages (UDP broadcast from a MASTER server instance)
        byte[] bRecvBuf = new byte[Heartbeat_Message_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, Heartbeat_Message_Size);
        try
        {
            m_HeartbeatReceiveSock.receive(RecvPacket);
        }
        catch(SocketTimeoutException Ex)
        {
             return;
        }
        catch(Exception Ex)
        {
            System.out.printf("Receive failed: %s\n", Ex.toString());
            return;
        }

        if (Heartbeat_Message_Size <= RecvPacket.getLength())
        {
            Heartbeat_Message_PDU Message = DeSerialize_HeartbeatMessage(bRecvBuf);
            switch(Message.eMessageType)
            {
                case HEARTBEAT:
                    if((Message.iSender_Random_ID + 1) == Message.iSender_Random_ID_Incremented)
                    { // Message is a genuine heartbeat message, not spurious message on heartbeat port (important check because broadcast is used, not a connection)
                        if(m_iServer_Random_ID != Message.iSender_Random_ID)
                        { // Ignore echos from self
                            String sStr = String.format("Heartbeat message received (from server ID:%d)",Message.iSender_Random_ID);
                            WriteEventLogEntry(sStr);
                            m_iHeartbeatMissedCount = 0;
                            jTextField_HeartbeatMissedCount.setText(Integer.toString(m_iHeartbeatMissedCount));
                            if(ServerState.MASTER == m_ServerState)
                            {
                                Transition_To_BACKUP_State(false);
                            }
                            else
                            {
                                m_Timer_Receive_Heartbeat.cancel();
                                Start_Timer_Receive_Heartbeat(); // Restart heartbeat timer so that it does not timeout

                                // A Master instance is present and this is a Backup instance so
                                // if this instance was freshly started it should request a database transfer from the Master instance
                                // because it may have been shutdown for some time and missed several database update propagations
                                if(true == m_bFreshlyStarted)
                                {   // Need to synchronise databases			
                                    Send_DatabaseTransferRequest(); // Request a database transfer from the Master instance
                                    m_bFreshlyStarted = false;
                                }
                            }
                        }
                    }
                    break;
                case REQUEST_FULL_DATABASE_TRANSFER:
                    if ((Message.iSender_Random_ID + 1) == Message.iSender_Random_ID_Incremented)
                    {   // Message is a genuine DB transfer request message, not spurious message on heartbeat port (important check because broadcast is used, not a connection)
                        if (m_iServer_Random_ID != Message.iSender_Random_ID)
                        {   // Ignore echos from self
                            String sStr = String.format("DB transfer request message received (from server ID:%d)", Message.iSender_Random_ID);
                            WriteEventLogEntry(sStr);
                            if (ServerState.MASTER == m_ServerState)
                            {	// Only the master instance responds to a DB transfer request							
                                Send_Database_ToBackupInstance();
                            }
                        }
                    }
                    break;
            }
        }
    }

    void Update_ServerState_Display()
    {
        jTextField_ServerState.setText(m_ServerState.toString());
        jTextField_HeartbeatMissedCount.setText(Integer.toString(m_iHeartbeatMissedCount));
    }
    
    ServerAdvertisement_Message_PDU DeSerialize_ServerAdvertisementMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
        ServerAdvertisement_Message_PDU result_Message = new ServerAdvertisement_Message_PDU();
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);            
        result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
        // Deserialisation of port number
        // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
        // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
        // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
        // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
        result_Message.iPort = (256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/;
        result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
        result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
        result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
        result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
        return result_Message;
    }
        
    Heartbeat_Message_PDU DeSerialize_HeartbeatMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a Heartbeat_Message_PDU structure
        Heartbeat_Message_PDU result_Message = new Heartbeat_Message_PDU();
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);            
        System.arraycopy(bRecdByteArray/*src*/, 8, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID_Incremented = ByteArraySize4_ToInt(bInt);            
        return result_Message;
    }

    Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
        Application_Message_PDU result_Message = new Application_Message_PDU();
        result_Message.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 4, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iSender_Random_ID = ByteArraySize4_ToInt(bInt);
        System.arraycopy(bRecdByteArray/*src*/, 8, bInt/*dest*/, 0, 4/*count*/); 
        result_Message.iCust_ID = ByteArraySize4_ToInt(bInt);
        result_Message.bCustNameValid = true;
        if (0 == bRecdByteArray[12])
        {
            result_Message.bCustNameValid = false;
        }
        result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
        System.arraycopy(bRecdByteArray/*src*/, 13, result_Message.cCustName/*dest*/, 0, MAX_CUST_NAME_LEN/*count*/); 
        return result_Message;
    }

    private byte[] Serialize_ServerAdvertisement_Message_PDU(ServerAdvertisement_Message_PDU Message)
    {   // Convert an ServerAdvertisement_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[ServerAdvertisement_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bByteArray[8] = (byte)Message.iAddressFamily;   // For C++ compatibility, value 2 = AF_INET

        // Serialisation of port number
        // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
        // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
        // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
        // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
        bByteArray[10] = (byte)(Message.iPort / 256);
        bByteArray[11] = (byte)(Message.iPort % 256);
        bByteArray[12] = Message.iIP_Address_Byte_1;
        bByteArray[13] = Message.iIP_Address_Byte_2;
        bByteArray[14] = Message.iIP_Address_Byte_3;
        bByteArray[15] = Message.iIP_Address_Byte_4;
        return bByteArray;
    }

    private byte[] Serialize_Heartbeat_Message_PDU(Heartbeat_Message_PDU Message)
    {   // Convert an Heartbeat_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Heartbeat_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bInt = IntToByteArray(Message.iSender_Random_ID_Incremented);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 8, 4/*count*/);
        return bByteArray;
    }

    private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
    {   // Convert an Application_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        byte[] bInt = IntToByteArray(Message.iSender_Random_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 4, 4/*count*/);
        bInt = IntToByteArray(Message.iCust_ID);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 8, 4/*count*/);
        bByteArray[12] = 0; // false (bCustNameValid)
        if(true == Message.bCustNameValid)
        {
            bByteArray[12] = 1; // true (bCustNameValid)
        }
        if (null != Message.cCustName)
        {
            System.arraycopy(Message.cCustName/*src*/, 0, bByteArray/*dest*/, 13, Message.cCustName.length/*count*/);
        }
        return bByteArray;
    }

    InetAddress GetLocalBroadcastAddress_From_LocalHostAddress(InetAddress LocalAddress) throws Exception
    {   // Assumes IPv4 address
        try
        {
            byte[] bLocalBroadcast = LocalAddress.getAddress(); // Initialise byte array with local host's address (e.g. 192.168.100.6)
            // Set the host portion of the address to the broadcast value, depending on the address class
            // Class A from 0.0.0.0 to 127.255.255.255   will be converted to the broadcast address x.255.255.255 
            // Class B from 128.0.0.0 to 191.255.255.255   will be converted to the broadcast address x.x.255.255  
            // Class C 192.0.0.0 to 223.255.255.255   will be converted to the broadcast address x.x.x.255 
            // Class D and E (224.0.0.0 to 225.255.255.254)   will be converted to the broadcast address 255.255.255.255 
            
            // For comparison purposes, the first byte of the address is converted to an int value (to ensure operations are performed 'unsigned')
            int iMostSignificantAddrsssByte = (int) bLocalBroadcast[0] & 0xFF;
            if(iMostSignificantAddrsssByte >= 0 && iMostSignificantAddrsssByte <= 127)
            {   // Class A address
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 128 && iMostSignificantAddrsssByte <= 191)
            {   // Class B address
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 192 && iMostSignificantAddrsssByte <= 223)
            {   // Class C address
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 224)
            {   // Class D / E address
                bLocalBroadcast[0] = (byte)255;
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            InetAddress LocalBroadcastAddress = InetAddress.getByAddress(bLocalBroadcast);
            return LocalBroadcastAddress;
        }
        catch (Exception Ex)
        {
            throw(Ex);
        }
    }
     
    String DottedIPAddressString_from_4ByteArray(byte[] bAddress)
    {
        String sStr = "";
        int iComponent = 0;
        if(4 != bAddress.length)
        {
            return sStr;    // If byte array is not 4 bytes return empty string
        }
        iComponent = bAddress[0] & 0xFF;
        sStr = Integer.toString(iComponent) + ".";
        iComponent = bAddress[1] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[2] & 0xFF;
        sStr += Integer.toString(iComponent) + ".";
        iComponent = bAddress[3] & 0xFF;            
        sStr += Integer.toString(iComponent);
        return sStr;
    }

    void CloseConnectionSocket()
    {
        try
        {
            m_ConnectSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }
    
    void CloseSockets()
    {
       CloseConnectionSocket();
        try
        {
            m_ListenSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_HeartbeatBroadcastSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_HeartbeatReceiveSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_ServerAdvertisementBroadcastSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_BroadcastPropagatedUpdateSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
        try
        {
            m_ReceivePropagatedUpdateSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not connected
        }
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
        if(null != m_Timer_Application)
        {
            m_Timer_Application.cancel();
        }
        if(null != m_Timer_Send_Heartbeat)
        {
            m_Timer_Send_Heartbeat.cancel();
        }
        if(null != m_Timer_Receive_Heartbeat)
        {
            m_Timer_Receive_Heartbeat.cancel();
        }
        if(null != m_Timer_Send_ServerAdvertisement)
        {
            m_Timer_Send_ServerAdvertisement.cancel();
        }
        CloseSockets();
        System.exit(0);
    }
        
    private static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    private int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }

    private String TruncateString(String sStr, int iMaxLength)
    {
        try
        {
            // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
            int iNullTerminatePosition = sStr.indexOf(0);
            if(-1 != iNullTerminatePosition)
            {
                sStr = sStr.substring(0, iNullTerminatePosition);
            }
            if(sStr.length()> iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.substring(0, iMaxLength);
            }
            return sStr;
        }
        catch (Exception Ex)
        {
            return "";
        }
    }
   
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        jTextArea_EventLog.append(sStr);       
    }

    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new Replication_Server_DB_Transfer().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
        if(jButton_ClearLog == e.getSource())
        {
            jTextArea_EventLog.setText("");
        }
    }
    
    private JPanel jPanel_Info;
    private JLabel jLabel_Info1;
    private JLabel jLabel_Info2;
    private JLabel jLabel_Info3;
    private JPanel jPanel_Data;
    private JLabel jLabel_CustomerID;
    private JTextArea jTextArea_CustomerID;
    private JLabel jLabel_CustomerName;
    private JTextArea jTextArea_CustomerName;
    private JPanel jPanel_State;
    private JLabel jLabel_ServerState;
    private JTextField jTextField_ServerState;
    private JLabel jLabel_ServerID;
    private JTextField jTextField_ServerID;
    private JLabel jLabel_HeartbeatMissedCount;
    private JTextField jTextField_HeartbeatMissedCount;
    private JPanel jPanel_Controls;
    private JButton jButton_ClearLog;
    private JButton jButton_Done;
    private JPanel jPanel_EventLog;
    private JTextArea jTextArea_EventLog;
    private JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_Info1 = new JLabel();
        jLabel_Info1.setText("Note: This component forms part of a demonstration of replication. As such it has only partial database");
        jLabel_Info2 = new JLabel();
        jLabel_Info2.setText("functionality. The data displayed initially is hardcoded, but can be updated to demonstrate update");
        jLabel_Info3 = new JLabel();
        jLabel_Info3.setText("propagation between service instances.");
        jPanel_Info = new JPanel();
        jPanel_Info.setPreferredSize(new Dimension(500, 100));
        jPanel_Info.setBorder(BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Info);
        jPanel_Info.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jLabel_Info1)
                .add(jLabel_Info2)
                .add(jLabel_Info3));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel1Layout.createSequentialGroup()
                    .add(jLabel_Info1)
                    .add(jLabel_Info2)
                    .add(jLabel_Info3)));

        jLabel_CustomerID = new JLabel();
        jLabel_CustomerID.setText("Customer ID");
        jTextArea_CustomerID = new JTextArea();
        jTextArea_CustomerID.setMaximumSize(new Dimension(80, 20));
        jLabel_CustomerName = new JLabel();
        jLabel_CustomerName.setText("Customer Name");
        jTextArea_CustomerName = new JTextArea();
        jTextArea_CustomerName.setMaximumSize(new Dimension(120, 20));
        jPanel_Data = new JPanel();
        jPanel_Data.setPreferredSize(new Dimension(250, 300));
        jPanel_Data.setBorder(BorderFactory.createTitledBorder("Data held by server instance"));

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_Data);
        jPanel_Data.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                    .add(jPanel2Layout.createSequentialGroup()
                        .add(jPanel2Layout.createParallelGroup()
                            .add(jLabel_CustomerID)
                            .add(jTextArea_CustomerID))
                        .addContainerGap(20, 20)
                        .add(jPanel2Layout.createParallelGroup()
                            .add(jLabel_CustomerName)
                            .add(jTextArea_CustomerName)))));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerID)
                    .addContainerGap(10, 10)
                    .add(jTextArea_CustomerID))
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jLabel_CustomerName)
                    .addContainerGap(10, 10)
                    .add(jTextArea_CustomerName)));            
            
        jLabel_ServerState = new JLabel();
        jLabel_ServerState.setText("Server instance state");
        jTextField_ServerState = new JTextField();
        jTextField_ServerState.setMaximumSize(new Dimension(80, 20));
        jLabel_ServerID = new JLabel();
        jLabel_ServerID.setText("Server ID");
        jTextField_ServerID = new JTextField();
        jTextField_ServerID.setMaximumSize(new Dimension(80, 20));
        jLabel_HeartbeatMissedCount = new JLabel();
        jLabel_HeartbeatMissedCount.setText("Heartbeat missed count");
        jTextField_HeartbeatMissedCount = new JTextField();
        jTextField_HeartbeatMissedCount.setMaximumSize(new Dimension(80, 20));
        
        jPanel_State = new JPanel();
        jPanel_State.setPreferredSize(new Dimension(250, 300));
        jPanel_State.setBorder(BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_State);
        jPanel_State.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_ServerState)
                    .addContainerGap(10, 10)
                    .add(jTextField_ServerState))
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_ServerID)
                    .addContainerGap(100, 100)
                    .add(jTextField_ServerID))
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jLabel_HeartbeatMissedCount)
                    .addContainerGap(10, 10)
                    .add(jTextField_HeartbeatMissedCount)));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jPanel3Layout.createParallelGroup()
                        .add(jLabel_ServerState)
                        .add(jTextField_ServerState))
                    .addContainerGap(10, 10)
                    .add(jPanel3Layout.createParallelGroup()
                        .add(jLabel_ServerID)
                        .add(jTextField_ServerID))
                    .addContainerGap(10, 10)           
                    .add(jPanel3Layout.createParallelGroup()
                        .add(jLabel_HeartbeatMissedCount)
                        .add(jTextField_HeartbeatMissedCount))));            
        
        jButton_ClearLog = new javax.swing.JButton();
        jButton_ClearLog.setText("Clear Log entries");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new javax.swing.JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls = new javax.swing.JPanel();
        jPanel_Controls.setPreferredSize(new Dimension(250, 200));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder(""));

        org.jdesktop.layout.GroupLayout jPanel4Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel4Layout);
        jPanel4Layout.setHorizontalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jButton_ClearLog)
            .add(jButton_Done));
        jPanel4Layout.setVerticalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel4Layout.createSequentialGroup()
                .add(jButton_ClearLog)
                .addContainerGap(10, 10)
                .add(jButton_Done)));
        jPanel4Layout.linkSize(new java.awt.Component[] {jButton_ClearLog, jButton_Done});

        jTextArea_EventLog = new javax.swing.JTextArea();
        jScrollPane_EventLog = new javax.swing.JScrollPane(jTextArea_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog = new javax.swing.JPanel();
        jPanel_EventLog.setPreferredSize(new Dimension(400, 400));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event Log (Diagnostic)"));
        
        org.jdesktop.layout.GroupLayout jPanel5Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel5Layout);
        jPanel5Layout.setHorizontalGroup(
            jPanel5Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));
        jPanel5Layout.setVerticalGroup(
            jPanel5Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jScrollPane_EventLog));        
         
        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(600, 400));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)            
                .add(jPanel_Info)
                .add(layout.createSequentialGroup()
                    .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)  
                        .add(jPanel_Data)
                        .add(jPanel_State)
                        .add(jPanel_Controls))
                    .add(layout.createParallelGroup()  
                        .add(jPanel_EventLog)))));
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
                .add(jPanel_Info)
                .add(layout.createParallelGroup()  
                    .add(layout.createSequentialGroup()
                        .add(jPanel_Data)
                        .addContainerGap(100, 100)
                        .add(jPanel_State)
                        .addContainerGap(100, 100)
                        .add(jPanel_Controls)
                        .addContainerGap(100, 100))
                    .add(layout.createSequentialGroup()
                        .add(jPanel_EventLog)))));

        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        setTitle("Replication SERVER (with DB transfer on startup)");
        pack();
    }                                             
}